'############################################################################
'# PFTH - This program implements a Forth interpreter for the P2.
'#
'# Copyright (c) 2012-2015 Dave Hein
'# MIT Licensed
'############################################################################
con
  _clkmode = xtal1+pll16x
  _clkfreq = 80_000_000

  Q = 16    ' Object Offset

  FLAG_IMMEDIATE = 1
  FLAG_CORE = $10 | $80
  FLAG_LIT  = $12 | $80
  FLAG_VAR  = $20 | $80
  FLAG_DEF  = $00 | $80
  FLAG_JMP  = $0A | $80
  FLAG_SEMI = FLAG_CORE | FLAG_IMMEDIATE

'*******************************************************************************
' This Spin code waits three seconds and then starts the Forth cog
'*******************************************************************************
pub main(argc, argv)
' waitcnt(clkfreq+cnt)
' coginit(cogid, @forth, @pfthconfig)
  
dat
{0018 000} pfthconfig              long      @xboot_1+Q            ' Initial word to execute
{001c 001}                         long      @stack+Q+16           ' Starting stack pointer
{0020 002}                         long      @stack+Q+16           ' Empty stack pointer value
{0024 003}                         long      @retstk+Q             ' Starting return pointer
{0028 004}                         long      @retstk+Q             ' Empty return pointer value
{002c 005} stack                   long      0[100]                ' Data stack
{01bc 069} retstk                  long      0[100]                ' Return stack

{034c 0cd}                         long      0[685]                ' Padding to get to $e00
                        
'*******************************************************************************
' pfth cog code
'*******************************************************************************
{0e00 000}                         org     0                
{0e00 000} forth
{0e00 000} parm
{0e00 000}   long $603c01e6  'parm                    mov     parm, parval
{0e04 001} parm1
{0e04 001}   long $c63c2800  'parm1                   rdlong  pc, parm
{0e08 002} parm2
{0e08 002}   long $107c0004  'parm2                   add     parm, #4
{0e0c 003} parm3
{0e0c 003}   long $c63fc000  'parm3                   rdlong  stackptr, parm
{0e10 004} parm4
{0e10 004}   long $107c0004  'parm4                   add     parm, #4
{0e14 005} temp
{0e14 005}   long $c63fc200  'temp                    rdlong  stackptr0, parm
{0e18 006} temp1
{0e18 006}   long $107c0004  'temp1                   add     parm, #4
{0e1c 007} temp2
{0e1c 007}   long $c63fc400  'temp2                   rdlong  returnptr, parm
{0e20 008} temp3
{0e20 008}   long $107c0004  'temp3                   add     parm, #4
{0e24 009} temp4
{0e24 009}   long $c63fc600  'temp4                   rdlong  returnptr0, parm
'                       mov     outa, outbit
'                       mov     dira, outbit
{0e28 00a}   long $d83c000f  '                        jmp     #innerloop              ' Begin execution

'*******************************************************************************
' Execute the words contained in the body of a word
' Changes parm, temp1
'*******************************************************************************
{0e2c 00b} execlistfunc
{0e2c 00b}   long $107c0004  'execlistfunc            add     parm, #4                ' Get body from XT
{0e30 00c} innerloopcall
{0e30 00c}   long $cb3c29e2  'innerloopcall           wrlong  pc, returnptr           ' Push PC to return stack
{0e34 00d}   long $107fc404  '                        add     returnptr, #4
{0e38 00e}   long $603c2800  '                        mov     pc, parm                ' Set new value for PC
                        
'*******************************************************************************
' Get an execution token from the location pointed to by the program counter
' Increment the program counter, fetch the code pointer and jump to it
' Changes parm, temp1, pc
'*******************************************************************************
{0e3c 00f} innerloop
{0e3c 00f}   long $c4bc0014  'innerloop               rdword  parm, pc                wz
{0e40 010}   long $d8280015  ' if_z                   jmp     #exitfunc
{0e44 011}   long $107c2802  '                        add     pc, #2
{0e48 012}   long $c43c0c00  '                        rdword  temp1, parm
{0e4c 013}   long $aa3ff406  '                        calld   ina, temp1
                        
{0e50 014} pc                      long    @xboot_1+Q             ' Program Counter

'*******************************************************************************
' Stop executing the current word, and return to the calling word
' No Changes              
'*******************************************************************************
{0e54 015} exitfunc
{0e54 015}   long $187fc404  'exitfunc                sub     returnptr, #4
{0e58 016}   long $c63c29e2  '                        rdlong  pc, returnptr
{0e5c 017}   long $d83c000f  '                        jmp     #innerloop

'*******************************************************************************
' Abort or quit execution, and return to the interpreter
' No Changes
'*******************************************************************************
{0e60 018} abortfunc
{0e60 018}   long $603fc1e1  'abortfunc               mov     stackptr, stackptr0
{0e64 019} quitfunc
{0e64 019}   long $603fc5e3  'quitfunc                mov     returnptr, returnptr0
{0e68 01a}   long $107fc404  '                        add     returnptr, #4           ' Use second entry return stack
{0e6c 01b}   long $c63c29e2  '                        rdlong  pc, returnptr      
{0e70 01c}   long $d83c000f  '                        jmp     #innerloop

'*******************************************************************************
' Push the value contained in the word's body onto the stack
' No changes
'*******************************************************************************
{0e74 01d} confunc
{0e74 01d}   long $107c0004  'confunc                 add     parm, #4
{0e78 01e}   long $c63c0200  '                        rdlong  parm1, parm
{0e7c 01f}   long $d83c01ad  '                        jmp     #push_jmp
                        
'*******************************************************************************
' Push the address of the word's body onto the stack
' Execute the words pointed to by the does pointer, if non-zero
' No changes
'*******************************************************************************
{0e80 020} varfunc
{0e80 020}   long $603c0200  'varfunc                 mov     parm1, parm
{0e84 021}   long $107c0204  '                        add     parm1, #4
{0e88 022}   long $da3c01aa  '                        call    #push1
                        
'*******************************************************************************
' Execute the words pointed to by the does pointer, if non-zero
' No changes
'*******************************************************************************
{0e8c 023} deferfunc
{0e8c 023}   long $107c0002  'deferfunc               add     parm, #2                ' DOES> pointer
{0e90 024}   long $c4bc0000  '                        rdword  parm, parm              wz
{0e94 025}   long $d828000f  '        if_z            jmp     #innerloop              ' Done with varfunc
{0e98 026}   long $d83c000c  '                        jmp     #innerloopcall          ' Execute DOES> code

'*******************************************************************************
' Execute the word on the stack
' Changes parm, temp1
'*******************************************************************************
{0e9c 027} executefunc
{0e9c 027}   long $187fc004  'executefunc             sub     stackptr, #4
{0ea0 028}   long $c63c01e0  '                        rdlong  parm, stackptr
{0ea4 029}   long $c63c0c00  '                        rdlong  temp1, parm
{0ea8 02a}   long $aa3ff406  '                        calld   ina, temp1                   ' Execute code

'*******************************************************************************
' Execute the PASM instruction on the TOS using the next value on the stack as
' the destination register data.  Return the result on the stack.
' Changes parm1, parm2
'*******************************************************************************
{0eac 02b} cogx1func
{0eac 02b}   long $da3c01b0  'cogx1func               call    #pop2
{0eb0 02c}   long $603c5e02  '                        mov     cogx1instr, parm2
{0eb4 02d}   long $997c5e01  '                        setd    cogx1instr, #parm1
{0eb8 02e}   long $00000000  '                        nop
{0ebc 02f} cogx1instr
{0ebc 02f}   long $00000000  'cogx1instr              nop
{0ec0 030}   long $d83c01ad  '                        jmp     #push_jmp

'*******************************************************************************
' Duplicate the top of stack
' Changes parm1
'*******************************************************************************
{0ec4 031} dupfunc
{0ec4 031}   long $da3c01b2  'dupfunc                 call    #pop1
{0ec8 032}   long $da3c01aa  '                        call    #push1
{0ecc 033}   long $d83c01ad  '                        jmp     #push_jmp

'*******************************************************************************
' Swap the top two items on the stack
' Changes parm1, parm2
'*******************************************************************************
{0ed0 034} swapfunc
{0ed0 034}   long $da3c01b0  'swapfunc                call    #pop2
{0ed4 035}   long $cb3c05e0  '                        wrlong  parm2, stackptr
{0ed8 036}   long $107fc004  '                        add     stackptr, #4
{0edc 037}   long $d83c01ad  '                        jmp     #push_jmp
                                                
'*******************************************************************************
' Get the next word from the input buffer using the delimiter from the stack
' Changes parm, parm1, parm2, temp1, temp2
'*******************************************************************************
{0ee0 038} wordfunc
{0ee0 038}   long $187fc004  'wordfunc                sub     stackptr, #4
{0ee4 039}   long $c63c01e0  '                        rdlong  parm, stackptr
{0ee8 03a}   long $da3c0165  '                        call    #word_del
{0eec 03b}   long $607c0c01  '                        mov     temp1, #1
{0ef0 03c}   long $067c0c0f  '                        shl     temp1, #15
{0ef4 03d}   long $183c0c02  '                        sub     temp1, parm2
{0ef8 03e}   long $187c0c01  '                        sub     temp1, #1
{0efc 03f}   long $cb3c0de0  '                        wrlong  temp1, stackptr
{0f00 040}   long $107fc004  '                        add     stackptr, #4
{0f04 041}   long $c93c0406  '                        wrbyte  parm2, temp1
{0f08 042}   long $25fc0400  '                        cmps    parm2, #0               wc, wz
{0f0c 043}   long $d838000f  '        if_c_or_z       jmp     #innerloop
{0f10 044} wordfunc1
{0f10 044}   long $107c0c01  'wordfunc1               add     temp1, #1
{0f14 045}   long $c23c0e01  '                        rdbyte  temp2, parm1
{0f18 046}   long $107c0201  '                        add     parm1, #1
{0f1c 047}   long $c93c0e06  '                        wrbyte  temp2, temp1
{0f20 048}   long $9cfc05fc  '                        djnz    parm2, @wordfunc1
{0f24 049}   long $d83c000f  '                        jmp     #innerloop

'*******************************************************************************
' Find the word specfied on the stack in the dictionary
' Changes parm1, parm2, temp4
'*******************************************************************************
{0f28 04a} findfunc
{0f28 04a}   long $da3c01b2  'findfunc                call    #pop1
{0f2c 04b}   long $603c1201  '                        mov     temp4, parm1
{0f30 04c}   long $107c0201  '                        add     parm1, #1
{0f34 04d}   long $c23c0409  '                        rdbyte  parm2, temp4
{0f38 04e}   long $da3c0172  '                        call    #findword
{0f3c 04f}   long $60bc0200  '                        mov     parm1, parm             wz
{0f40 050}   long $d828005a  '        if_z            jmp     #findfunc1
{0f44 051}   long $da3c01bb  '                        call    #link2xt
{0f48 052}   long $da3c01aa  '                        call    #push1
{0f4c 053}   long $107c0002  '                        add     parm, #2                ' Point to flag byte
{0f50 054}   long $c23c0200  '                        rdbyte  parm1, parm
{0f54 055}   long $527c0201  '                        and     parm1, #1               ' Check immediate bit
{0f58 056}   long $067c0201  '                        shl     parm1, #1
{0f5c 057}   long $187c0201  '                        sub     parm1, #1               ' Return 1 if set, -1 if not
{0f60 058}   long $da3c01aa  '                        call    #push1
{0f64 059}   long $d83c000f  '                        jmp     #innerloop
{0f68 05a} findfunc1
{0f68 05a}   long $603c0209  'findfunc1               mov     parm1, temp4
{0f6c 05b}   long $da3c01aa  '                        call    #push1
{0f70 05c}   long $607c0200  '                        mov     parm1, #0
{0f74 05d}   long $da3c01aa  '                        call    #push1
{0f78 05e}   long $d83c000f  '                        jmp     #innerloop

'*******************************************************************************
' Send the character from the stack to the output port
' Changes parm
'*******************************************************************************
{0f7c 05f} emitfunc
{0f7c 05f}   long $da3c01b2  'emitfunc                call    #pop1
{0f80 060}   long $603c0001  '                        mov     parm, parm1
{0f84 061}   long $da3c0143  '                        call    #putch
{0f88 062}   long $d83c000f  '                        jmp     #innerloop

'*******************************************************************************
' Get a character from the input port and put it on the stack
' Changes parm, parm1
'*******************************************************************************
{0f8c 063} getcharfunc
{0f8c 063}   long $da3c0130  'getcharfunc             call    #getch
{0f90 064}   long $603c0200  '                        mov     parm1, parm
{0f94 065}   long $d83c01ad  '                        jmp     #push_jmp

'*******************************************************************************
' Get a character from the files stored in memory and put it on the stack
' Changes parm1
'*******************************************************************************
{0f98 066} getfcharfunc
{0f98 066}   long $c23c03e4  'getfcharfunc            rdbyte  parm1, infileptr
{0f9c 067}   long $107fc801  '                        add     infileptr, #1
{0fa0 068}   long $d83c01ad  '                        jmp     #push_jmp

'*******************************************************************************
' Get an address and value from the stack, and store the value at the address
' No changes              
'*******************************************************************************
{0fa4 069} storefunc
{0fa4 069}   long $da3c01b0  'storefunc               call    #pop2
{0fa8 06a}   long $cb3c0202  '                        wrlong  parm1, parm2
{0fac 06b}   long $d83c000f  '                        jmp     #innerloop

'*******************************************************************************
' Fetch a value from the address specified on the stack, and put it on the stack
' Changes parm1              
'*******************************************************************************
{0fb0 06c} fetchfunc
{0fb0 06c}   long $da3c01b2  'fetchfunc               call    #pop1
{0fb4 06d}   long $c63c0201  '                        rdlong  parm1, parm1
{0fb8 06e}   long $d83c01ad  '                        jmp     #push_jmp

'*******************************************************************************
' Get an address and word from the stack, and store the word at the address
' No changes              
'*******************************************************************************
{0fbc 06f} wstorefunc
{0fbc 06f}   long $da3c01b0  'wstorefunc              call    #pop2
{0fc0 070}   long $ca3c0202  '                        wrword  parm1, parm2
{0fc4 071}   long $d83c000f  '                        jmp     #innerloop

'*******************************************************************************
' Fetch a word from the address specified on the stack, and put it on the stack
' Changes parm1              
'*******************************************************************************
{0fc8 072} wfetchfunc
{0fc8 072}   long $da3c01b2  'wfetchfunc              call    #pop1
{0fcc 073}   long $c43c0201  '                        rdword  parm1, parm1
{0fd0 074}   long $d83c01ad  '                        jmp     #push_jmp
              
'*******************************************************************************
' Get an address and byte from the stack, and store the byte at the address
' No changes              
'*******************************************************************************
{0fd4 075} cstorefunc
{0fd4 075}   long $da3c01b0  'cstorefunc              call    #pop2
{0fd8 076}   long $c93c0202  '                        wrbyte  parm1, parm2
{0fdc 077}   long $d83c000f  '                        jmp     #innerloop

'*******************************************************************************
' Fetch a byte from the address specified on the stack, and put it on the stack
' Changes parm1              
'*******************************************************************************
{0fe0 078} cfetchfunc
{0fe0 078}   long $da3c01b2  'cfetchfunc              call    #pop1
{0fe4 079}   long $c23c0201  '                        rdbyte  parm1, parm1
{0fe8 07a}   long $d83c01ad  '                        jmp     #push_jmp
              
'*******************************************************************************
' Add two values from the stack, and write the result back to the stack
' Changes parm1, parm2              
'*******************************************************************************
{0fec 07b} plusfunc
{0fec 07b}   long $da3c01b0  'plusfunc                call    #pop2
{0ff0 07c}   long $103c0202  '                        add     parm1, parm2
{0ff4 07d}   long $d83c01ad  '                        jmp     #push_jmp
              
'*******************************************************************************
' Subtract two values from the stack, and write the result back to the stack
' Changes parm1, parm2              
'*******************************************************************************
{0ff8 07e} minusfunc
{0ff8 07e}   long $da3c01b0  'minusfunc               call    #pop2
{0ffc 07f}   long $183c0202  '                        sub     parm1, parm2
{1000 080}   long $d83c01ad  '                        jmp     #push_jmp
              
'*******************************************************************************
' Multiply two values from the stack, and write the result back to the stack
' Changes parm1, parm2              
'*******************************************************************************
{1004 081} multfunc
{1004 081}   long $da3c01b0  'multfunc                call    #pop2
{1008 082}   long $da3c01c2  '                        call    #multiply
{100c 083}   long $d83c01ad  '                        jmp     #push_jmp
              
'*******************************************************************************
' Divide two values from the stack, and write the result back to the stack
' Changes parm1, parm2              
'*******************************************************************************
{1010 084} dividefunc
{1010 084}   long $da3c01b0  'dividefunc              call    #pop2
{1014 085}   long $da3c01ca  '                        call    #divide
{1018 086}   long $603c0202  '                        mov     parm1, parm2
{101c 087}   long $a37c0601  '                        test    parm3, #1               wc
{1020 088}   long $66300201  '        if_c            neg     parm1, parm1
{1024 089}   long $d83c01ad  '                        jmp     #push_jmp
              
'*******************************************************************************
' Compute the modulus from two values from the stack, and write the result back
' to the stack
' Changes parm1, parm2              
'*******************************************************************************
{1028 08a} modfunc
{1028 08a}   long $da3c01b0  'modfunc                 call    #pop2
{102c 08b}   long $da3c01ca  '                        call    #divide
{1030 08c}   long $a37c0602  '                        test    parm3, #2               wc
{1034 08d}   long $66300201  '        if_c            neg     parm1, parm1
{1038 08e}   long $d83c01ad  '                        jmp     #push_jmp
              
'*******************************************************************************
' Compare two values from the stack to determine if the second one is less than
' the first one, and write the result back to the stack
' Changes parm1, parm2              
'*******************************************************************************
{103c 08f} lessfunc
{103c 08f}   long $da3c01b0  'lessfunc                call    #pop2
{1040 090}   long $253c0202  '                        cmps    parm1, parm2            wc
{1044 091}   long $66700201  '        if_c            neg     parm1, #1
{1048 092}   long $604c0200  '        if_nc           mov     parm1, #0
{104c 093}   long $d83c01ad  '                        jmp     #push_jmp
              
'*******************************************************************************
' Compare two values from the stack to determine if they are equal, and write
' the result back to the stack
' Changes parm1, parm2              
'*******************************************************************************
{1050 094} equalfunc
{1050 094}   long $da3c01b0  'equalfunc               call    #pop2
{1054 095}   long $20bc0202  '                        cmp     parm1, parm2            wz
{1058 096}   long $66680201  '        if_z            neg     parm1, #1
{105c 097}   long $60540200  '        if_nz           mov     parm1, #0
{1060 098}   long $d83c01ad  '                        jmp     #push_jmp
              
'*******************************************************************************
' Compare two values from the stack to determine if the second one is greater
' than the first one, and write the result back to the stack
' Changes parm1, parm2              
'*******************************************************************************
{1064 099} greaterfunc
{1064 099}   long $da3c01b0  'greaterfunc             call    #pop2
{1068 09a}   long $25bc0202  '                        cmps    parm1, parm2            wc, wz
{106c 09b}   long $66440201  '        if_nz_and_nc    neg     parm1, #1
{1070 09c}   long $60780200  '        if_z_or_c       mov     parm1, #0
{1074 09d}   long $d83c01ad  '                        jmp     #push_jmp

'*******************************************************************************
' Compute the logical AND of two values from the stack, and write the result
' back to the stack
' Changes parm1, parm2              
'*******************************************************************************
{1078 09e} andfunc
{1078 09e}   long $da3c01b0  'andfunc                 call    #pop2
{107c 09f}   long $523c0202  '                        and     parm1, parm2
{1080 0a0}   long $d83c01ad  '                        jmp     #push_jmp
              
'*******************************************************************************
' Compute the logical OR of two values from the stack, and write the result
' back to the stack
' Changes parm1, parm2              
'*******************************************************************************
{1084 0a1} orfunc
{1084 0a1}   long $da3c01b0  'orfunc                  call    #pop2
{1088 0a2}   long $543c0202  '                        or      parm1, parm2
{108c 0a3}   long $d83c01ad  '                        jmp     #push_jmp
              
'*******************************************************************************
' Compute the logical XOR of two values from the stack, and write the result
' back to the stack
' Changes parm1, parm2              
'*******************************************************************************
{1090 0a4} xorfunc
{1090 0a4}   long $da3c01b0  'xorfunc                 call    #pop2
{1094 0a5}   long $563c0202  '                        xor     parm1, parm2
{1098 0a6}   long $d83c01ad  '                        jmp     #push_jmp
              
'*******************************************************************************
' Right-shift the second value on the stack by the number of bits specified by
' the first value on the stack, and write the result to the stack
' Changes parm1, parm2              
'*******************************************************************************
{109c 0a7} rshiftfunc
{109c 0a7}   long $da3c01b0  'rshiftfunc              call    #pop2
{10a0 0a8}   long $043c0202  '                        shr     parm1, parm2
{10a4 0a9}   long $d83c01ad  '                        jmp     #push_jmp
              
'*******************************************************************************
' Left-shift the second value on the stack by the number of bits specified by
' the first value on the stack, and write the result to the stack
' Changes parm1, parm2              
'*******************************************************************************
{10a8 0aa} lshiftfunc
{10a8 0aa}   long $da3c01b0  'lshiftfunc              call    #pop2
{10ac 0ab}   long $063c0202  '                        shl     parm1, parm2
{10b0 0ac}   long $d83c01ad  '                        jmp     #push_jmp

'*******************************************************************************
' Push the stack depth to the stack
' Changes parm1              
'*******************************************************************************
{10b4 0ad} depthfunc
{10b4 0ad}   long $603c03e0  'depthfunc               mov     parm1, stackptr
{10b8 0ae}   long $183c03e1  '                        sub     parm1, stackptr0
{10bc 0af}   long $0c7c0202  '                        sar     parm1, #2
{10c0 0b0}   long $d83c01ad  '                        jmp     #push_jmp

'*******************************************************************************
' Drop the top value from the stack
' No changes
'*******************************************************************************
{10c4 0b1} dropfunc
{10c4 0b1}   long $187fc004  'dropfunc                sub     stackptr, #4
{10c8 0b2}   long $d83c000f  '                        jmp     #innerloop

'*******************************************************************************
' Use the value on top of the stack as an index to another value in the stack,
' and write its value to the stack
' No changes
'*******************************************************************************
{10cc 0b3} pickfunc
{10cc 0b3}   long $da3c01b2  'pickfunc                call    #pop1
{10d0 0b4}   long $da3c01b5  '                        call    #indexstack
{10d4 0b5}   long $d83c01ad  '                        jmp     #push_jmp

'*******************************************************************************
' Use the value on top of the stack as and index to remove another value from
' the stack, and place it at the top of the stack.
' Changes temp1, temp2, temp3, temp4
'*******************************************************************************
{10d8 0b6} rollfunc
{10d8 0b6}   long $da3c01b2  'rollfunc                call    #pop1
{10dc 0b7}   long $21fc0200  '                        cmp     parm1, #0               wc, wz
{10e0 0b8}   long $d838000f  '        if_c_or_z       jmp     #innerloop
{10e4 0b9}   long $603c1001  '                        mov     temp3, parm1
{10e8 0ba}   long $da3c01b5  '                        call    #indexstack
{10ec 0bb}   long $603c0e06  '                        mov     temp2, temp1
{10f0 0bc} rollfunc1
{10f0 0bc}   long $107c0e04  'rollfunc1               add     temp2, #4
{10f4 0bd}   long $c63c1207  '                        rdlong  temp4, temp2
{10f8 0be}   long $cb3c1206  '                        wrlong  temp4, temp1
{10fc 0bf}   long $107c0c04  '                        add     temp1, #4
{1100 0c0}   long $9cfc11fc  '                        djnz    temp3, @rollfunc1
{1104 0c1}   long $cb3c0206  '                        wrlong  parm1, temp1
{1108 0c2}   long $d83c000f  '                        jmp     #innerloop

'*******************************************************************************
' Pop the value from the top of the stack, and push it onto the return stack.
' No changes
'*******************************************************************************
{110c 0c3} torfunc
{110c 0c3}   long $da3c01b2  'torfunc                 call    #pop1
{1110 0c4}   long $cb3c03e2  '                        wrlong  parm1, returnptr
{1114 0c5}   long $107fc404  '                        add     returnptr, #4
{1118 0c6}   long $d83c000f  '                        jmp     #innerloop

'*******************************************************************************
' Pop the value from the top of the return stack and push it to the stack.
' Changes parm1
'*******************************************************************************
{111c 0c7} fromrfunc
{111c 0c7}   long $187fc404  'fromrfunc               sub     returnptr, #4
{1120 0c8}   long $c63c03e2  '                        rdlong  parm1, returnptr
{1124 0c9}   long $d83c01ad  '                        jmp     #push_jmp

'*******************************************************************************
' Push the value on the stack pointed to by the PC and increment the PC
' Changes parm1
'*******************************************************************************
{1128 0ca} _litfunc
{1128 0ca}   long $c43c0214  '_litfunc                rdword  parm1, pc
{112c 0cb}   long $107c2802  '                        add     pc, #2
{1130 0cc}   long $d83c01ad  '                        jmp     #push_jmp

'*******************************************************************************
' Convert the string described by the address and length on the top of the
' stack to a hex number, and push it to the stack
' Changes parm1
'*******************************************************************************
{1134 0cd} _gethexfunc
{1134 0cd}   long $da3c01b0  '_gethexfunc             call    #pop2
{1138 0ce}   long $da3c019c  '                        call    #gethex
{113c 0cf}   long $603c0200  '                        mov     parm1, parm
{1140 0d0}   long $d83c01ad  '                        jmp     #push_jmp

'*******************************************************************************
' Create a variable, and add it to the dictionary
' Changes parm3
'*******************************************************************************
{1144 0d1} createfunc
{1144 0d1}   long $607c0620  'createfunc              mov     parm3, #varfunc
{1148 0d2}   long $607c08a0  '                        mov     parm4, #FLAG_VAR
{114c 0d3}   long $da3c010f  '                        call    #create
{1150 0d4}   long $d83c000f  '                        jmp     #innerloop

'*******************************************************************************
' Create an executable word, and add it to the dictionary.  Set the compile
' state to -1
' Changes parm3, temp1
'*******************************************************************************
{1154 0d5} colonfunc
{1154 0d5}   long $607c060b  'colonfunc               mov     parm3, #execlistfunc
{1158 0d6}   long $607c0880  '                        mov     parm4, #FLAG_DEF
{115c 0d7}   long $da3c010f  '                        call    #create
{1160 0d8}   long $d828000f  '        if_z            jmp     #innerloop
{1164 0d9}   long $667c0c01  '                        neg     temp1, #1
{1168 0da}   long $cb3c0dda  '                        wrlong  temp1, a_state
{116c 0db}   long $d83c000f  '                        jmp     #innerloop

'*******************************************************************************
' Compile a zero into memory indicating the end of an executable word, and set
' the compile flag to zero
' Changes temp1, temp2
'*******************************************************************************
{1170 0dc} semicolonfunc
{1170 0dc}   long $607c0c00  'semicolonfunc           mov     temp1, #0
{1174 0dd}   long $cb3c0dda  '                        wrlong  temp1, a_state
{1178 0de}   long $c63c0fdb  '                        rdlong  temp2, a_dp
{117c 0df}   long $ca3c0c07  '                        wrword  temp1, temp2
{1180 0e0}   long $107c0e02  '                        add     temp2, #2
{1184 0e1}   long $cb3c0fdb  '                        wrlong  temp2, a_dp
{1188 0e2}   long $d83c000f  '                        jmp     #innerloop

'*******************************************************************************
' Fetch a value from the specified cog address, and put it on the stack
' the compile flag to zero
' Changes parm1
'*******************************************************************************
{118c 0e3} cogfetchfunc
{118c 0e3}   long $da3c01b2  'cogfetchfunc            call    #pop1
{1190 0e4}   long $983dcc01  '                        sets    cogfetch1, parm1
{1194 0e5}   long $00000000  '                        nop
{1198 0e6} cogfetch1
{1198 0e6}   long $603c0200  'cogfetch1               mov     parm1, 0-0
{119c 0e7}   long $d83c01ad  '                        jmp     #push_jmp

'*******************************************************************************
' Get a cog address and value from the stack, and store the value at the address
' the compile flag to zero
' Changes parm1, parm2
'*******************************************************************************
{11a0 0e8} cogstorefunc
{11a0 0e8}   long $da3c01b0  'cogstorefunc            call    #pop2
{11a4 0e9}   long $993dd602  '                        setd    cogstore1, parm2
{11a8 0ea}   long $00000000  '                        nop
{11ac 0eb} cogstore1
{11ac 0eb}   long $603c0001  'cogstore1               mov     0-0, parm1
{11b0 0ec}   long $d83c000f  '                        jmp     #innerloop


'*******************************************************************************
' Print out an 8-digit hex number to the output port.
' Changes parm
'*******************************************************************************
{11b4 0ed} dotxfunc
{11b4 0ed}   long $607c0024  'dotxfunc                mov     parm, #"$"
{11b8 0ee}   long $da3c0143  '                        call    #putch
{11bc 0ef}   long $da3c01b2  '                        call    #pop1
{11c0 0f0}   long $da3c0193  '                        call    #printhex
{11c4 0f1}   long $607c0020  '                        mov     parm, #" "
{11c8 0f2}   long $da3c0143  '                        call    #putch
{11cc 0f3}   long $d83c000f  '                        jmp     #innerloop

'*******************************************************************************
' If top of stack is zero, jump to address contained in location at current PC.
' Otherwise, increment the PC
' Changes parm1
'*******************************************************************************
{11d0 0f4} _jzfunc
{11d0 0f4}   long $da3c01b2  '_jzfunc                 call    #pop1
{11d4 0f5}   long $c4282814  '        if_z            rdword  pc, pc
{11d8 0f6}   long $10542802  '        if_nz           add     pc, #2
{11dc 0f7}   long $d83c000f  '                        jmp     #innerloop

'*******************************************************************************
' Copy bytes from the source to the destination
' Changes parm1
'*******************************************************************************
{11e0 0f8} cmovefunc
{11e0 0f8}   long $187fc004  'cmovefunc               sub     stackptr, #4
{11e4 0f9}   long $c63c07e0  '                        rdlong  parm3, stackptr
{11e8 0fa}   long $da3c01b0  '                        call    #pop2
{11ec 0fb}   long $25fc0600  '                        cmps    parm3, #0               wz, wc
{11f0 0fc}   long $d838000f  '        if_c_or_z       jmp     #innerloop
{11f4 0fd} cmovefunc1
{11f4 0fd}   long $c23c0c01  'cmovefunc1              rdbyte  temp1, parm1
{11f8 0fe}   long $107c0201  '                        add     parm1, #1
{11fc 0ff}   long $c93c0c02  '                        wrbyte  temp1, parm2
{1200 100}   long $107c0401  '                        add     parm2, #1
{1204 101}   long $9cfc07fc  '                        djnz    parm3, @cmovefunc1
{1208 102}   long $d83c000f  '                        jmp     #innerloop

'*******************************************************************************
' Perform the increment and compare for the loop word
' Changes parm1, parm2, parm3
'*******************************************************************************
{120c 103} _loopfunc
{120c 103}   long $da3c01b2  '_loopfunc               call    #pop1                   ' Get increment
{1210 104}   long $187fc408  '                        sub     returnptr, #8
{1214 105}   long $c63c07e2  '                        rdlong  parm3, returnptr        ' Get upper limit
{1218 106}   long $107fc404  '                        add     returnptr, #4
{121c 107}   long $c63c05e2  '                        rdlong  parm2, returnptr        ' Get index
{1220 108}   long $103c0202  '                        add     parm1, parm2            ' index + increment
{1224 109}   long $cb3c03e2  '                        wrlong  parm1, returnptr        ' Push index back
{1228 10a}   long $107fc404  '                        add     returnptr, #4
{122c 10b}   long $253c0203  '                        cmps    parm1, parm3            wc
{1230 10c}   long $664c0201  '        if_nc           neg     parm1, #1
{1234 10d}   long $60700200  '        if_c            mov     parm1, #0
{1238 10e}   long $d83c01ad  '                        jmp     #push_jmp
                        
'*******************************************************************************
' The following code implements the basic functions used by the kernel words
'*******************************************************************************

'*******************************************************************************
' Create a word entry in the dictionary
' Changes parm, parm1, parm2, temp1, temp2              
'*******************************************************************************
{123c 10f} create
{123c 10f}   long $607c0020  'create                  mov     parm, #" "
{1240 110}   long $da3c0165  '                        call    #word_del
{1244 111}   long $d628002d  '        if_z            ret
{1248 112}   long $c63c0ddb  '                        rdlong  temp1, a_dp             ' Align DP
{124c 113}   long $107c0c03  '                        add     temp1, #3
{1250 114}   long $523c0de5  '                        and     temp1, minus4
{1254 115}   long $c63c0fd9  '                        rdlong  temp2, a_last
{1258 116}   long $ca3c0e06  '                        wrword  temp2, temp1            ' Write the link pointer
{125c 117}   long $cb3c0dd9  '                        wrlong  temp1, a_last           ' Update LAST
{1260 118}   long $107c0c02  '                        add     temp1, #2
                        
{1264 119}   long $c93c0806  '                        wrbyte  parm4, temp1            ' Write the flag
{1268 11a}   long $107c0c01  '                        add     temp1, #1
{126c 11b}   long $c93c0406  '                        wrbyte  parm2, temp1            ' Write the length
{1270 11c}   long $107c0c01  '                        add     temp1, #1
{1274 11d}   long $25fc0400  '                        cmps    parm2, #0               wc, wz
{1278 11e}   long $d8380124  '        if_c_or_z       jmp     #create_done
{127c 11f} create1
{127c 11f}   long $c23c0e01  'create1                 rdbyte  temp2, parm1            ' Copy the name
{1280 120}   long $107c0201  '                        add     parm1, #1
{1284 121}   long $c93c0e06  '                        wrbyte  temp2, temp1
{1288 122}   long $10fc0c01  '                        add     temp1, #1               wz
{128c 123}   long $9cfc05fc  '                        djnz    parm2, @create1

{1290 124} create_done
{1290 124}   long $607c0e00  'create_done             mov     temp2, #0               ' Pad with 0's to align
{1294 125} create2
{1294 125}   long $a2fc0c03  'create2                 test    temp1, #3               wz
{1298 126}   long $d828012a  '        if_z            jmp     #create_aligned
{129c 127}   long $c93c0e06  '                        wrbyte  temp2, temp1
{12a0 128}   long $107c0c01  '                        add     temp1, #1
{12a4 129}   long $d83c0125  '                        jmp     #create2

{12a8 12a} create_aligned
{12a8 12a}   long $ca3c0606  'create_aligned          wrword  parm3, temp1            ' Write the code pointer
{12ac 12b}   long $107c0c02  '                        add     temp1, #2
{12b0 12c}   long $ca3c0e06  '                        wrword  temp2, temp1            ' Write the DOES> pointer             
{12b4 12d}   long $10fc0c02  '                        add     temp1, #2               wz ' Clear zero flag
                                  
{12b8 12e}   long $cb3c0ddb  '                        wrlong  temp1, a_dp
{12bc 12f} create_ret
{12bc 12f}   long $d63c002d  'create_ret              ret

'*******************************************************************************
' Get one character from the input port.
' Input none
' Changes parm, temp, temp1, temp2
' Output parm
'*******************************************************************************
{12c0 130} getch
{12c0 130}   long $603c01fa  'getch                   mov     parm, ina
{12c4 131}   long $52bc0141  '                        and     parm, inbit             wz
{12c8 132}   long $d8140130  '        if_nz           jmp     #getch
{12cc 133}   long $d63c0e1e  '                        getcnt  temp2
{12d0 134}   long $603c0b42  '                        mov     temp, bitcycles
{12d4 135}   long $047c0a01  '                        shr     temp, #1
{12d8 136}   long $103c0e05  '                        add     temp2, temp
{12dc 137}   long $607c0c0a  '                        mov     temp1, #10
{12e0 138} getch1
{12e0 138}   long $a83c0f42  'getch1                  waitcnt temp2, bitcycles
{12e4 139}   long $603c0bfa  '                        mov     temp, ina
{12e8 13a}   long $523c0b41  '                        and     temp, inbit
{12ec 13b}   long $007c0001  '                        ror     parm, #1
{12f0 13c}   long $543c0005  '                        or      parm, temp
{12f4 13d}   long $9cfc0dfb  '                        djnz    temp1, @getch1
{12f8 13e}   long $007c0017  '                        ror     parm, #31 - 8
{12fc 13f}   long $527c00ff  '                        and     parm, #255
{1300 140} getch_ret
{1300 140}   long $d63c002d  'getch_ret               ret

{1304 141} inbit                   long    $80000000
{1308 142} bitcycles               long    80_000_000 / 115_200

'*******************************************************************************
' Send one character to the output port.
' Input parm
' Changes parm, temp1, temp2
' Output none             
'*******************************************************************************
{130c 143} putch
{130c 143}   long $c6bc0ddd  'putch                   rdlong  temp1, a_verbose       wz
{1310 144}   long $d628002d  '        if_z            ret
{1314 145} putch0
{1314 145}   long $547c0100  'putch0                  or      parm, #$100
{1318 146}   long $067c0001  '                        shl     parm, #1
{131c 147}   long $607c0c0a  '                        mov     temp1, #10
{1320 148}   long $d63c0e1e  '                        getcnt  temp2
{1324 149}   long $103c0f42  '                        add     temp2, bitcycles
{1328 14a} putch1
{1328 14a}   long $057c0001  'putch1                  shr     parm, #1               wc
{132c 14b}   long $5433f950  '        if_c            or      outa, outbit
{1330 14c}   long $500ff950  '        if_nc           andn    outa, outbit
{1334 14d}   long $a83c0f42  '                        waitcnt temp2, bitcycles
{1338 14e}   long $9cfc0dfc  '                        djnz    temp1, @putch1
{133c 14f} putch_ret
{133c 14f}   long $d63c002d  'putch_ret               ret

{1340 150} outbit                  long    $40000000

'*******************************************************************************
' Skip the specified character in the input buffer
' Input parm
' Changes temp, temp1
' Output none
'*******************************************************************************
{1344 151} skipchar
{1344 151}   long $253c0c07  'skipchar                cmps    temp1, temp2            wc
{1348 152}   long $d60c002d  '        if_nc           ret
{134c 153}   long $c63c0bdc  '                        rdlong  temp, a_tib
{1350 154}   long $103c0a06  '                        add     temp, temp1
{1354 155}   long $c23c0a05  '                        rdbyte  temp, temp
{1358 156}   long $20bc0a00  '                        cmp     temp, parm              wz
{135c 157}   long $d614002d  '        if_nz           ret
{1360 158}   long $107c0c01  '                        add     temp1, #1
{1364 159}   long $d83c0151  '                        jmp     #skipchar
{1368 15a} skipchar_ret
{1368 15a}   long $d63c002d  'skipchar_ret            ret

'*******************************************************************************
' Find the next occurance of the specified character in the input buffer
' Input parm
' Changes temp, temp1
' Output none
'*******************************************************************************
{136c 15b} findchar
{136c 15b}   long $253c0c07  'findchar                cmps    temp1, temp2            wc
{1370 15c}   long $d60c002d  '        if_nc           ret
{1374 15d}   long $c63c0bdc  '                        rdlong  temp, a_tib
{1378 15e}   long $103c0a06  '                        add     temp, temp1
{137c 15f}   long $c23c0a05  '                        rdbyte  temp, temp
{1380 160}   long $20bc0a00  '                        cmp     temp, parm              wz
{1384 161}   long $d628002d  '        if_z            ret
{1388 162}   long $107c0c01  '                        add     temp1, #1
{138c 163}   long $d83c015b  '                        jmp     #findchar
{1390 164} findchar_ret
{1390 164}   long $d63c002d  'findchar_ret            ret

'*******************************************************************************
' Find the next word in the input buffer delimited by the specified character
' Input parm
' Changes parm1, parm2, temp1, temp2
' Output none
'*******************************************************************************
{1394 165} word_del
{1394 165}   long $c63c0dde  '                        rdlong  temp1, a_inputidx
{1398 166}   long $c63c0fdf  '                        rdlong  temp2, a_inputlen
{139c 167}   long $da3c0151  '                        call    #skipchar
{13a0 168}   long $603c0206  '                        mov     parm1, temp1
{13a4 169}   long $da3c015b  '                        call    #findchar
{13a8 16a}   long $603c0406  '                        mov     parm2, temp1
{13ac 16b}   long $18bc0401  '                        sub     parm2, parm1            wz
{13b0 16c}   long $c63c0bdc  '                        rdlong  temp, a_tib
{13b4 16d}   long $103c0205  '                        add     parm1, temp
{13b8 16e}   long $253c0c07  '                        cmps    temp1, temp2            wc
{13bc 16f}   long $10700c01  '        if_c            add     temp1, #1
{13c0 170}   long $cb3c0dde  '                        wrlong  temp1, a_inputidx
{13c4 171} word_del_ret
{13c4 171}   long $d63c002d  'word_del_ret            ret

'*******************************************************************************
' Find the specified word in the dictionary
' Input parm1, parm2
' Changes parm, parm3, parm4
' Output parm
'*******************************************************************************
{13c8 172} findword
{13c8 172}   long $c6bc01d9  'findword                rdlong  parm, a_last            wz
{13cc 173}   long $d628002d  '        if_z            ret
{13d0 174} findword1
{13d0 174}   long $603c0600  'findword1               mov     parm3, parm
{13d4 175}   long $107c0603  '                        add     parm3, #3
{13d8 176}   long $c23c0803  '                        rdbyte  parm4, parm3
{13dc 177}   long $107c0601  '                        add     parm3, #1
{13e0 178}   long $da3c017d  '                        call    #compare
{13e4 179}   long $d628002d  '        if_z            ret
{13e8 17a}   long $c4bc0000  '                        rdword  parm, parm              wz
{13ec 17b}   long $d8140174  '        if_nz           jmp     #findword1
{13f0 17c} findword_ret
{13f0 17c}   long $d63c002d  'findword_ret            ret

'*******************************************************************************
' Do a case insensitive comparison of two character strings
' Input parm1, parm2, parm3, parm4
' Changes parm3, parm4, temp, temp1, temp2
' Outut Z
'*******************************************************************************
{13f4 17d} compare
{13f4 17d}   long $25fc0401  'compare                 cmps    parm2, #1               wc, wz
{13f8 17e}   long $d630002d  '        if_c            ret
{13fc 17f}   long $20bc0404  '                        cmp     parm2, parm4            wz
{1400 180}   long $d614002d  '        if_nz           ret
{1404 181}   long $603c0a01  '                        mov     temp, parm1
{1408 182} compare1
{1408 182}   long $c23c0c05  'compare1                rdbyte  temp1, temp
{140c 183}   long $da3c018d  '                        call    #toupper
{1410 184}   long $603c0e06  '                        mov     temp2, temp1
{1414 185}   long $c23c0c03  '                        rdbyte  temp1, parm3
{1418 186}   long $da3c018d  '                        call    #toupper
{141c 187}   long $20bc0c07  '                        cmp     temp1, temp2            wz
{1420 188}   long $d614002d  '        if_nz           ret
{1424 189}   long $107c0a01  '                        add     temp, #1
{1428 18a}   long $107c0601  '                        add     parm3, #1
{142c 18b}   long $9cfc09f7  '                        djnz    parm4, @compare1
{1430 18c} compare_ret
{1430 18c}   long $d63c002d  'compare_ret             ret

'*******************************************************************************
' Convert a character to uppercase
' Input temp1
' Changes temp1
' Ouput temp1
'*******************************************************************************
{1434 18d} toupper
{1434 18d}   long $217c0c61  'toupper                 cmp     temp1, #"a"             wc
{1438 18e}   long $d630002d  '        if_c            ret
{143c 18f}   long $21fc0c7a  '                        cmp     temp1, #"z"             wc, wz
{1440 190}   long $d604002d  '        if_nc_and_nz    ret
{1444 191}   long $187c0c20  '                        sub     temp1, #"a" - "A"
{1448 192} toupper_ret
{1448 192}   long $d63c002d  'toupper_ret             ret              


'*******************************************************************************
' Print an 8-digit hex value to the output port
' Input parm1
' Changes parm, parm1, parm2
' Output none
'*******************************************************************************
{144c 193} printhex
{144c 193}   long $607c0408  'printhex                mov     parm2, #8
{1450 194} printhex1
{1450 194}   long $027c0204  'printhex1               rol     parm1, #4
{1454 195}   long $607c000f  '                        mov     parm, #15
{1458 196}   long $523c0001  '                        and     parm, parm1
{145c 197}   long $103c01d8  '                        add     parm, a_hexstr
{1460 198}   long $c23c0000  '                        rdbyte  parm, parm
{1464 199}   long $da3c0143  '                        call    #putch
{1468 19a}   long $9cfc05fa  '                        djnz    parm2, @printhex1
{146c 19b} printhex_ret
{146c 19b}   long $d63c002d  'printhex_ret            ret


'*******************************************************************************
' Convert a string to a hex number
' Input parm1, parm2
' Changes parm, temp, temp1, temp2
' Output parm
'*******************************************************************************
{1470 19c} gethex
{1470 19c}   long $607c0000  'gethex                  mov     parm, #0
{1474 19d}   long $25fc0400  '                        cmps    parm2, #0               wc, wz
{1478 19e}   long $d638002d  '        if_c_or_z       ret
{147c 19f}   long $603c0c01  '                        mov     temp1, parm1
{1480 1a0}   long $603c0e02  '                        mov     temp2, parm2
{1484 1a1} gethex1
{1484 1a1}   long $c23c0a06  'gethex1                 rdbyte  temp, temp1
{1488 1a2}   long $107c0c01  '                        add     temp1, #1
{148c 1a3}   long $187c0a30  '                        sub     temp, #"0"
{1490 1a4}   long $257c0a0a  '                        cmps    temp, #10               wc
{1494 1a5}   long $184c0a27  '        if_nc           sub     temp, #"a"-"0"-10
{1498 1a6}   long $067c0004  '                        shl     parm, #4
{149c 1a7}   long $103c0005  '                        add     parm, temp
{14a0 1a8}   long $9cfc0ff9  '                        djnz    temp2, @gethex1
{14a4 1a9} gethex_ret
{14a4 1a9}   long $d63c002d  'gethex_ret              ret

'*******************************************************************************
' Push a value onto the data stack
' Input parm1
' No changes
' Output none
'*******************************************************************************
{14a8 1aa} push1
{14a8 1aa}   long $cb3c03e0  'push1                   wrlong  parm1, stackptr
{14ac 1ab}   long $107fc004  '                        add     stackptr, #4
{14b0 1ac} push1_ret
{14b0 1ac}   long $d63c002d  'push1_ret               ret

'*******************************************************************************
' Push a value onto the data stack and jump to the innerloop
' Input parm1
' No changes
' Output none
'*******************************************************************************
{14b4 1ad} push_jmp
{14b4 1ad}   long $cb3c03e0  'push_jmp                wrlong  parm1, stackptr
{14b8 1ae}   long $107fc004  '                        add     stackptr, #4
{14bc 1af}   long $d83c000f  '                        jmp     #innerloop

'*******************************************************************************
' Pop two values off of the data stack
' Input none
' Changes parm1, parm2
' Output parm1, parm2
'*******************************************************************************
{14c0 1b0} pop2
{14c0 1b0}   long $187fc004  'pop2                    sub     stackptr, #4
{14c4 1b1}   long $c63c05e0  '                        rdlong  parm2, stackptr

'*******************************************************************************
' Pop one value off of the data stack
' Input none
' Changes parm1
' Ouput parm1
'*******************************************************************************
{14c8 1b2} pop1
{14c8 1b2}   long $187fc004  'pop1                    sub     stackptr, #4
{14cc 1b3}   long $c6bc03e0  '                        rdlong  parm1, stackptr         wz
{14d0 1b4} pop1_ret
{14d0 1b4} pop2_ret
{14d0 1b4}   long $d63c002d  'pop2_ret                ret

'*******************************************************************************
' Read a value on the stack based on an index number
' Changes parm1, temp1
'*******************************************************************************
{14d4 1b5} indexstack
{14d4 1b5}   long $663c0c01  'indexstack              neg     temp1, parm1
{14d8 1b6}   long $067c0c02  '                        shl     temp1, #2
{14dc 1b7}   long $187c0c04  '                        sub     temp1, #4
{14e0 1b8}   long $103c0de0  '                        add     temp1, stackptr
{14e4 1b9}   long $c63c0206  '                        rdlong  parm1, temp1
{14e8 1ba} indexstack_ret
{14e8 1ba}   long $d63c002d  'indexstack_ret          ret

'*******************************************************************************
' Compute the XT from the address of the link
' Input:   parm1
' Output:  parm1
' Changes: temp1
'*******************************************************************************
{14ec 1bb} link2xt
{14ec 1bb}   long $603c0c01  'link2xt                 mov     temp1, parm1
{14f0 1bc}   long $107c0c03  '                        add     temp1, #3
{14f4 1bd}   long $c23c0206  '                        rdbyte  parm1, temp1            ' Get name length
{14f8 1be}   long $103c0206  '                        add     parm1, temp1
{14fc 1bf}   long $107c0204  '                        add     parm1, #4
{1500 1c0}   long $523c03e5  '                        and     parm1, minus4           ' Align
{1504 1c1} link2xt_ret
{1504 1c1}   long $d63c002d  'link2xt_ret             ret

'*******************************************************************************
' Multiply two 32-bit numbers
' Changes parm2, temp1, temp2
'*******************************************************************************
{1508 1c2} multiply
{1508 1c2}   long $607c0c00  'multiply                mov     temp1, #0
{150c 1c3}   long $607c0e20  '                        mov     temp2, #32
{1510 1c4}   long $057c0201  '                        shr     parm1, #1               wc
{1514 1c5} mmul
{1514 1c5}   long $11300c02  'mmul    if_c            add     temp1, parm2            wc
{1518 1c6}   long $097c0c01  '                        rcr     temp1, #1               wc
{151c 1c7}   long $097c0201  '                        rcr     parm1, #1               wc
{1520 1c8}   long $9cfc0ffd  '                        djnz    temp2, @mmul
{1524 1c9} multiply_ret
{1524 1c9}   long $d63c002d  'multiply_ret            ret

'*******************************************************************************
' Divide two 32-bit numbers producing a quotient and a remainder
' Changes parm1, parm2, parm3, temp1, temp2
'*******************************************************************************
{1528 1ca} divide
{1528 1ca}   long $607c0e20  'divide                  mov     temp2, #32
{152c 1cb}   long $607c0c00  '                        mov     temp1, #0
{1530 1cc}   long $653c0201  '                        abs     parm1, parm1            wc
{1534 1cd}   long $587c0603  '                        muxc    parm3, #%11
{1538 1ce}   long $65bc0402  '                        abs     parm2, parm2            wc,wz
{153c 1cf}   long $56700601  '        if_c            xor     parm3, #%01
'        if_nz           jmp     #mdiv
'                        mov     parm1, #0
'                        jmp     divide_ret
{1540 1d0} mdiv
{1540 1d0}   long $05fc0401  'mdiv                    shr     parm2, #1               wc,wz
{1544 1d1}   long $087c0c01  '                        rcr     temp1, #1
{1548 1d2}   long $9cd40ffe  '        if_nz           djnz    temp2, @mdiv
{154c 1d3} mdiv2
{154c 1d3}   long $2f3c0206  'mdiv2                   cmpsub  parm1, temp1            wc
{1550 1d4}   long $0a7c0401  '                        rcl     parm2, #1
{1554 1d5}   long $047c0c01  '                        shr     temp1, #1
{1558 1d6}   long $9cfc0ffd  '                        djnz    temp2, @mdiv2
{155c 1d7} divide_ret
{155c 1d7}   long $d63c002d  'divide_ret              ret

'*******************************************************************************
' These are working registers.  The parm registers are generally used to pass
' parameters from one routine to another, and the temp registers are used as
' temporary storage within a routine.
'*******************************************************************************

'*******************************************************************************
' Addresses of variables in the dictionary, and the hex table
'*******************************************************************************
{1560 1d8} a_hexstr                long      @hexstr+Q
{1564 1d9} a_last                  long      @last+Q
{1568 1da} a_state                 long      @state+Q
{156c 1db} a_dp                    long      @dp+Q
{1570 1dc} a_tib                   long      @tib+Q
{1574 1dd} a_verbose               long      @verbose+Q
{1578 1de} a_inputidx              long      @greaterin+Q
{157c 1df} a_inputlen              long      @poundtib+Q

'*******************************************************************************
' The data and return stack pointers, and their base addresses
'*******************************************************************************
{1580 1e0} stackptr                long      0
{1584 1e1} stackptr0               long      0
{1588 1e2} returnptr               long      0
{158c 1e3} returnptr0              long      0

'*******************************************************************************
' The input file pointer used during initialization
'*******************************************************************************
{1590 1e4} infileptr               long      @infile+Q

'*******************************************************************************
' Constants
'*******************************************************************************
{1594 1e5} minus4                  long      -4
{1598 1e6} parval                  long      @pfthconfig+Q

{159c 1e7}                         fit       $1f0

'*******************************************************************************
' Input buffer and hex table
'*******************************************************************************
{159c 1e7} hexstr                  byte      "0123456789abcdef"
{15ac 1eb} inputbuf                byte      0[200]

'*******************************************************************************
' This is the beginning of the dictionary.  The kernel words are specified below
'*******************************************************************************
{1674 21d} exit_L        word      0
{1676 21d}               byte      FLAG_CORE, 4, "exit"
{167c 21f}               long
{167c 21f} exit_X        word      exitfunc, 0

{1680 220} quit_L        word      @exit_L+Q
{1682 220}               byte      FLAG_CORE, 4, "quit"
{1688 222}               long
{1688 222} quit_X        word      quitfunc, 0

{168c 223} abort_L       word      @quit_L+Q
{168e 223}               byte      FLAG_CORE, 5, "abort"
{1698 226}               long
{1698 226} abort_X       word      abortfunc, 0

{169c 227} execute_L     word      @abort_L+Q
{169e 227}               byte      FLAG_CORE, 7, "execute"
{16a8 22a}               long
{16a8 22a} execute_X     word      executefunc, 0

{16ac 22b} word_L        word      @execute_L+Q
{16ae 22b}               byte      FLAG_CORE, 4, "word"
{16b4 22d}               long
{16b4 22d} word_X        word      wordfunc, 0

{16b8 22e} find_L        word      @word_L+Q
{16ba 22e}               byte      FLAG_CORE, 4, "find"
{16c0 230}               long
{16c0 230} find_X        word      findfunc, 0

{16c4 231} getchar_L     word      @find_L+Q
{16c6 231}               byte      FLAG_CORE, 7, "getchar"
{16d0 234}               long
{16d0 234} getchar_X     word      getcharfunc, 0

{16d4 235} getfchar_L    word      @getchar_L+Q
{16d6 235}               byte      FLAG_CORE, 8, "getfchar"
{16e0 238}               long
{16e0 238} getfchar_X    word      getfcharfunc, 0

{16e4 239} key_L         word      @getfchar_L+Q
{16e6 239}               byte      FLAG_CORE, 3, "key"
{16ec 23b}               long
{16ec 23b} key_X         word      deferfunc, @key_B+Q
'key_B         word      @getchar_X+Q, 0
{16f0 23c} key_B         word      @getfchar_X+Q, 0

{16f4 23d} create_L      word      @key_L+Q
{16f6 23d}               byte      FLAG_CORE, 6, "create"
{1700 240}               long
{1700 240} create_X      word      createfunc, 0

{1704 241} _lit_L        word      @create_L+Q
{1706 241}               byte      FLAG_LIT, 4, "_lit"
{170c 243}               long
{170c 243} _lit_X        word      _litfunc, 0

{1710 244} _gethex_L     word      @_lit_L+Q
{1712 244}               byte      FLAG_CORE, 7, "_gethex"
{171c 247}               long
{171c 247} _gethex_X     word      _gethexfunc, 0

{1720 248} emit_L        word      @_gethex_L+Q
{1722 248}               byte      FLAG_CORE, 4, "emit"
{1728 24a}               long
{1728 24a} emit_X        word      emitfunc, 0

{172c 24b} store_L       word      @emit_L+Q
{172e 24b}               byte      FLAG_CORE, 1, "!"
{1734 24d}               long
{1734 24d} store_X       word      storefunc, 0

{1738 24e} fetch_L       word      @store_L+Q
{173a 24e}               byte      FLAG_CORE, 1, "@"
{1740 250}               long
{1740 250} fetch_X       word      fetchfunc, 0

{1744 251} wstore_L      word      @fetch_L+Q
{1746 251}               byte      FLAG_CORE, 2, "w!"
{174c 253}               long
{174c 253} wstore_X      word      wstorefunc, 0

{1750 254} wfetch_L      word      @wstore_L+Q
{1752 254}               byte      FLAG_CORE, 2, "w@"
{1758 256}               long
{1758 256} wfetch_X      word      wfetchfunc, 0

{175c 257} cstore_L      word      @wfetch_L+Q
{175e 257}               byte      FLAG_CORE, 2, "c!"
{1764 259}               long
{1764 259} cstore_X      word      cstorefunc, 0

{1768 25a} cfetch_L      word      @cstore_L+Q
{176a 25a}               byte      FLAG_CORE, 2, "c@"
{1770 25c}               long
{1770 25c} cfetch_X      word      cfetchfunc, 0

{1774 25d} plus_L        word      @cfetch_L+Q
{1776 25d}               byte      FLAG_CORE, 1, "+"
{177c 25f}               long
{177c 25f} plus_X        word      plusfunc, 0

{1780 260} minus_L       word      @plus_L+Q
{1782 260}               byte      FLAG_CORE, 1, "-"
{1788 262}               long
{1788 262} minus_X       word      minusfunc, 0

{178c 263} multiply_L    word      @minus_L+Q
{178e 263}               byte      FLAG_CORE, 1, "*"
{1794 265}               long
{1794 265} multiply_X    word      multfunc, 0

{1798 266} divide_L      word      @multiply_L+Q
{179a 266}               byte      FLAG_CORE, 1, "/"
{17a0 268}               long
{17a0 268} divide_X      word      dividefunc, 0

{17a4 269} mod_L         word      @divide_L+Q
{17a6 269}               byte      FLAG_CORE, 3, "mod"
{17ac 26b}               long
{17ac 26b} mod_X         word      modfunc, 0

{17b0 26c} and_L         word      @mod_L+Q
{17b2 26c}               byte      FLAG_CORE, 3, "and"
{17b8 26e}               long
{17b8 26e} and_X         word      andfunc, 0

{17bc 26f} or_L          word      @and_L+Q
{17be 26f}               byte      FLAG_CORE, 2, "or"
{17c4 271}               long
{17c4 271} or_X          word      orfunc, 0

{17c8 272} xor_L         word      @or_L+Q
{17ca 272}               byte      FLAG_CORE, 3, "xor"
{17d0 274}               long
{17d0 274} xor_X         word      xorfunc, 0

{17d4 275} less_L        word      @xor_L+Q
{17d6 275}               byte      FLAG_CORE, 1, "<"
{17dc 277}               long
{17dc 277} less_X        word      lessfunc, 0

{17e0 278} equal_L       word      @less_L+Q
{17e2 278}               byte      FLAG_CORE, 1, "="
{17e8 27a}               long
{17e8 27a} equal_X       word      equalfunc, 0

{17ec 27b} greater_L     word      @equal_L+Q
{17ee 27b}               byte      FLAG_CORE, 1, ">"
{17f4 27d}               long
{17f4 27d} greater_X     word      greaterfunc, 0

{17f8 27e} rshift_L      word      @greater_L+Q
{17fa 27e}               byte      FLAG_CORE, 6, "rshift"
{1804 281}               long
{1804 281} rshift_X      word      rshiftfunc, 0

{1808 282} lshift_L      word      @rshift_L+Q
{180a 282}               byte      FLAG_CORE, 6, "lshift"
{1814 285}               long
{1814 285} lshift_X      word      lshiftfunc, 0

{1818 286} depth_L       word      @lshift_L+Q
{181a 286}               byte      FLAG_CORE, 5, "depth"
{1824 289}               long
{1824 289} depth_X       word      depthfunc, 0

{1828 28a} tib_L         word      @depth_L+Q
{182a 28a}               byte      FLAG_VAR, 3, "tib"
{1830 28c}               long
{1830 28c} tib_X         word      varfunc, @tib+Q+4
{1834 28d} tib           long      @inputbuf+Q
{1838 28e}               word      @fetch_X+Q, 0
{183c 28f}               long

{183c 28f} poundtib_L    word      @tib_L+Q
{183e 28f}               byte      FLAG_VAR, 4, "#tib"
{1844 291}               long
{1844 291} poundtib_X    word      varfunc, 0
{1848 292} poundtib      long      0

{184c 293} greaterin_L   word      @poundtib_L+Q
{184e 293}               byte      FLAG_VAR, 3, ">in"
{1854 295}               long
{1854 295} greaterin_X   word      varfunc, 0
{1858 296} greaterin     long      0

{185c 297} dp_L          word      @greaterin_L+Q
{185e 297}               byte      FLAG_VAR, 2, "dp"
{1864 299}               long
{1864 299} dp_X          word      varfunc, 0
{1868 29a} dp            long      @_here+Q

{186c 29b} last_L        word      @dp_L+Q
{186e 29b}               byte      FLAG_VAR, 4, "last"
{1874 29d}               long
{1874 29d} last_X        word      varfunc, 0
{1878 29e} last          long      @_last+Q

{187c 29f} state_L       word      @last_L+Q
{187e 29f}               byte      FLAG_VAR, 5, "state"
{1888 2a2}               long
{1888 2a2} state_X       word      varfunc, 0
{188c 2a3} state         long      0

{1890 2a4} base_L        word      @state_L+Q
{1892 2a4}               byte      FLAG_VAR, 4, "base"
{1898 2a6}               long
{1898 2a6} base_X        word      varfunc, 0
{189c 2a7} base          long      16

{18a0 2a8} verbose_L     word      @base_L+Q
{18a2 2a8}               byte      FLAG_VAR, 7, "verbose"
{18ac 2ab}               long
{18ac 2ab} verbose_X     word      varfunc, 0
{18b0 2ac} verbose       long      0

{18b4 2ad} forth_L       word      @verbose_L+Q
{18b6 2ad}               byte      FLAG_VAR, 5, "forth"
{18c0 2b0}               long
{18c0 2b0} forth_X       word      varfunc, 0
{18c4 2b1}               long      @forth+Q

{18c8 2b2} drop_L        word      @forth_L+Q
{18ca 2b2}               byte      FLAG_CORE, 4, "drop"
{18d0 2b4}               long
{18d0 2b4} drop_X        word      dropfunc, 0

{18d4 2b5} dup_L         word      @drop_L+Q
{18d6 2b5}               byte      FLAG_CORE, 3, "dup"
{18dc 2b7}               long
{18dc 2b7} dup_X         word      dupfunc, 0

{18e0 2b8} swap_L        word      @dup_L+Q
{18e2 2b8}               byte      FLAG_CORE, 4, "swap"
{18e8 2ba}               long
{18e8 2ba} swap_X        word      swapfunc, 0

{18ec 2bb} pick_L        word      @swap_L+Q
{18ee 2bb}               byte      FLAG_CORE, 4, "pick"
{18f4 2bd}               long
{18f4 2bd} pick_X        word      pickfunc, 0

{18f8 2be} roll_L        word      @pick_L+Q
{18fa 2be}               byte      FLAG_CORE, 4, "roll"
{1900 2c0}               long
{1900 2c0} roll_X        word      rollfunc, 0

{1904 2c1} tor_L         word      @roll_L+Q
{1906 2c1}               byte      FLAG_CORE, 2, ">r"
{190c 2c3}               long
{190c 2c3} tor_X         word      torfunc, 0

{1910 2c4} fromr_L       word      @tor_L+Q
{1912 2c4}               byte      FLAG_CORE, 2, "r>"
{1918 2c6}               long
{1918 2c6} fromr_X       word      fromrfunc, 0

{191c 2c7} colon_L       word      @fromr_L+Q
{191e 2c7}               byte      FLAG_CORE, 1, ":"
{1924 2c9}               long
{1924 2c9} colon_X       word      colonfunc, 0

{1928 2ca} semicolon_L   word      @colon_L+Q
{192a 2ca}               byte      FLAG_SEMI, 1, ";"
{1930 2cc}               long
{1930 2cc} semicolon_X   word      semicolonfunc, 0 

{1934 2cd} cogfetch_L    word      @semicolon_L+Q
{1936 2cd}               byte      FLAG_CORE, 4, "cog@"
{193c 2cf}               long
{193c 2cf} cogfetch_X    word      cogfetchfunc, 0

{1940 2d0} cogstore_L    word      @cogfetch_L+Q
{1942 2d0}               byte      FLAG_CORE, 4, "cog!"
{1948 2d2}               long
{1948 2d2} cogstore_X    word      cogstorefunc, 0

{194c 2d3} cogx1_L       word      @cogstore_L+Q
{194e 2d3}               byte      FLAG_CORE, 5, "cogx1"
{1958 2d6}               long
{1958 2d6} cogx1_X       word      cogx1func, 0

{195c 2d7} _jz_L         word      @cogx1_L+Q
{195e 2d7}               byte      FLAG_JMP, 3, "_jz"
{1964 2d9}               long
{1964 2d9} _jz_X         word      _jzfunc, 0

{1968 2da} cmove_L       word      @_jz_L+Q
{196a 2da}               byte      FLAG_CORE, 5, "cmove"
{1974 2dd}               long
{1974 2dd} cmove_X       word      cmovefunc, 0

{1978 2de} dotx_L        word      @cmove_L+Q
{197a 2de}               byte      FLAG_CORE, 2, ".x"
{1980 2e0}               long
{1980 2e0} dotx_X        word      dotxfunc, 0

'*******************************************************************************
' SPI/SD Variables
'*******************************************************************************
{1984 2e1} spi_vars_L    word      @dotx_L+Q
{1986 2e1}               byte      FLAG_VAR, 8, "spi_vars"
{1990 2e4}               long
{1990 2e4} spi_vars_X    word      varfunc, 0
{1994 2e5} spi_vars      long      0       ' SPI_engine_cog
{1998 2e6}               long      0       ' SPI_command
{199c 2e7}               long      0       ' SPI_block_index
{19a0 2e8}               long      0       ' SPI_buffer_address
{19a4 2e9}               long      0       ' SD_rootdir
{19a8 2ea}               long      0       ' SD_filesystem
{19ac 2eb}               long      0       ' SD_clustershift
{19b0 2ec}               long      0       ' SD_dataregion
{19b4 2ed}               long      0       ' SD_fat1
{19b8 2ee}               long      0       ' SD_sectorsperfat
{19bc 2ef}               long      0       ' SD_currdir

{19c0 2f0} argc_L        word      @spi_vars_L+Q
{19c2 2f0}               byte      FLAG_VAR, 4, "argc"
{19c8 2f2}               long
{19c8 2f2} argc_X        word      confunc, 0
{19cc 2f3} argc_B        long      0

{19d0 2f4} argv_L        word      @argc_L+Q
{19d2 2f4}               byte      FLAG_VAR, 4, "argv"
{19d8 2f6}               long
{19d8 2f6} argv_X        word      confunc, 0
{19dc 2f7} argv_B        long      0

{19e0 2f8} hostcwd_L     word      @argv_L+Q
{19e2 2f8}               byte      FLAG_VAR, 7, "hostcwd"
{19ec 2fb}               long
{19ec 2fb} hostcwd_X     word      confunc, 0
{19f0 2fc} hostcwd_B     long      0

'*******************************************************************************
' A small number of compiled words follow below.  These are used by the boot
' interpreter.
'*******************************************************************************
              ' : here dp @ ;
{19f4 2fd} here_L        word      @hostcwd_L+Q
{19f6 2fd}               byte      FLAG_DEF, 4, "here"
{19fc 2ff}               long
{19fc 2ff} here_X        word      execlistfunc, 0
{1a00 300}               word      @dp_X+Q, @fetch_X+Q, 0
{1a08 302}               long
              
              ' : allot dp @ + dp ! ;
{1a08 302} allot_L       word      @here_L+Q
{1a0a 302}               byte      FLAG_DEF, 5, "allot"
{1a14 305}               long
{1a14 305} allot_X       word      execlistfunc, 0
{1a18 306}               word      @dp_X+Q, @fetch_X+Q, @plus_X+Q, @dp_X+Q, @store_X+Q, 0
{1a24 309}               long

              ' : , here ! 4 allot ;
{1a24 309} comma_L       word      @allot_L+Q
{1a26 309}               byte      FLAG_DEF, 1, ","
{1a2c 30b}               long
{1a2c 30b} comma_X       word      execlistfunc, 0
{1a30 30c}               word      @here_X+Q, @store_X+Q, @_lit_X+Q, 4, @allot_X+Q, 0
{1a3c 30f}               long

              ' : _jmp r> @ >r ;
{1a3c 30f} _jmp_L        word      @comma_L+Q
{1a3e 30f}               byte      FLAG_JMP, 4, "_jmp"
{1a44 311}               long
{1a44 311} _jmp_X        word      execlistfunc, 0
{1a48 312}               word      @fromr_X+Q, @wfetch_X+Q, @tor_X+Q, 0
{1a50 314}               long

              ' : count 0 pick 1 + 1 roll c@ ;
{1a50 314} count_L       word      @_jmp_L+Q
{1a52 314}               byte      FLAG_DEF, 5, "count"
{1a5c 317}               long
{1a5c 317} count_X       word      execlistfunc, 0
{1a60 318}               word      @_lit_X+Q, 0, @pick_X+Q, @_lit_X+Q, 1, @plus_X+Q, @_lit_X+Q, 1, @roll_X+Q, @cfetch_X+Q, 0
{1a78 31e}               long

              ' : accept ( addr size -- num ) \ Accept a string from the input source
{1a78 31e} accept_L      word      @count_L+Q
{1a7a 31e}               byte      FLAG_DEF, 6, "accept"
{1a84 321}               long
{1a84 321} accept_X      word      execlistfunc, 0
              ' >r dup
{1a88 322}               word      @tor_X+Q, @dup_X+Q
              ' r> dup 1 < _jz _accept4
{1a8c 323} accept_1      word      @fromr_X+Q, @dup_X+Q, @_lit_X+Q, 1, @less_X+Q, @_jz_X+Q, @accept_4+Q
              ' drop swap - exit
{1a9a 326}               word      @drop_X+Q, @swap_X+Q, @minus_X+Q, @exit_X+Q
              ' >r key
{1aa2 328} accept_4      word      @tor_X+Q, @key_X+Q
              ' dup 0d = over 0a = or
{1aa6 329}               word      @dup_X+Q, @_lit_X+Q, $0d, @equal_X+Q, @_lit_X+Q, 1, @pick_X+Q, @_lit_X+Q, $0a, @equal_X+Q, @or_X+Q
              ' _jz _accept2
{1abc 32f}               word      @_jz_X+Q, @accept_2+Q
              ' cr drop swap -
{1ac0 330}               word      @_lit_X+Q, 13, @emit_X+Q, @_lit_X+Q, 10, @emit_X+Q, @drop_X+Q, @swap_X+Q, @minus_X+Q
              ' r> drop exit
{1ad2 334}               word      @fromr_X+Q, @drop_X+Q, @exit_X+Q
              ' dup 8 = _jz _accept3
{1ad8 336} accept_2      word      @dup_X+Q, @_lit_X+Q, 8, @equal_X+Q, @_jz_X+Q, @accept_3+Q
              ' drop over over - _jz _accept1
{1ae4 339}               word      @drop_X+Q, @_lit_X+Q, 1, @pick_X+Q, @_lit_X+Q, 1, @pick_X+Q, @minus_X+Q, @_jz_X+Q, @accept_1+Q
              ' 1 - r> 1 + >r
{1af8 33e}               word      @_lit_X+Q, 1, @minus_X+Q, @fromr_X+Q, @_lit_X+Q, 1, @plus_X+Q, @tor_X+Q
              ' 8 emit bl emit 8 emit _jmp _accept1
{1b08 342}               word      @_lit_X+Q, 8, @emit_X+Q, @_lit_X+Q, 32, @emit_X+Q, @_lit_X+Q, 8, @emit_X+Q, @_jmp_X+Q, @accept_1+Q
              ' dup emit over c! 1 +
{1b1e 347} accept_3      word      @dup_X+Q, @emit_X+Q, @_lit_X+Q, 1, @pick_X+Q, @cstore_X+Q, @_lit_X+Q, 1, @plus_X+Q
              ' r> 1 - >r _jmp _accept1
{1b30 34c}               word      @fromr_X+Q, @_lit_X+Q, 1, @minus_X+Q, @tor_X+Q, @_jmp_X+Q, @accept_1+Q, 0
{1b40 350}               long
              
              ' : refill tib 200 accept #tib ! 0 >in ! ;
{1b40 350} refill_L      word      @accept_L+Q
{1b42 350}               byte      FLAG_DEF, 6, "refill"
{1b4c 353}               long
{1b4c 353} refill_X      word      execlistfunc, 0
{1b50 354}               word      @tib_X+Q, @_lit_X+Q, 200, @accept_X+Q, @poundtib_X+Q, @store_X+Q, @_lit_X+Q, 0, @greaterin_X+Q, @store_X+Q, 0
{1b68 35a}               long

              ' : compile, here w! 2 allot ;
{1b68 35a} compcomma_L   word      @refill_L+Q
{1b6a 35a}               byte      FLAG_DEF, 8, "compile,"
{1b74 35d}               long
{1b74 35d} compcomma_X   word      execlistfunc, 0
{1b78 35e}               word      @here_X+Q, @wstore_X+Q, @_lit_X+Q, 2, @allot_X+Q, 0
{1b84 361}               long
              
'*******************************************************************************
' The boot interpreter follows below.
'*******************************************************************************
              ' : xboot ( This word runs a simple interpreter )
{1b84 361} xboot_L       word      @compcomma_L+Q
{1b86 361}               byte      FLAG_DEF, 5, "xboot"
{1b90 364}               long
{1b90 364} xboot_X       word      execlistfunc, 0

              ' 20 word 0 pick c@ _jz _xboot2 ( Get word, refill if empty )
{1b94 365} xboot_1       word      @_lit_X+Q, $20, @word_X+Q, @_lit_X+Q, 0, @pick_X+Q, @cfetch_X+Q, @_jz_X+Q, @xboot_2+Q
              
              ' find 0 pick _jz _xboot3 ( Find word, get number if not found )
{1ba6 369}               word      @find_X+Q, @_lit_X+Q, 0, @pick_X+Q, @_jz_X+Q, @xboot_3+Q
              
              ' state @ = _jz _xboot4 ( Go execute if not compile mode or immediate )
{1bb2 36c}               word      @state_X+Q, @fetch_X+Q, @equal_X+Q, @_jz_X+Q, @xboot_4+Q
              
              ' compile, _jmp _xboot1 ( Otherwise, compile and loop again )
{1bbc 36f}               word       @compcomma_X+Q, @_jmp_X+Q, @xboot_1+Q
              
              ' execute _jmp _xboot1 ( Execute and loop again )
{1bc2 370} xboot_4       word      @execute_X+Q, @_jmp_X+Q, @xboot_1+Q

              ' drop count _gethex ( Get number )
{1bc8 372} xboot_3       word      @drop_X+Q, @count_X+Q, @_gethex_X+Q

              ' state @ _jz _xboot1 ( Loop again if not compile mode )
{1bce 373}               word      @state_X+Q, @fetch_X+Q, @_jz_X+Q, @xboot_1+Q
              
              ' ['] _lit , , _jmp _xboot1 ( Otherwise, compile number and loop again )
{1bd6 375}               word       @_lit_X+Q, @_lit_X+Q, @compcomma_X+Q, @compcomma_X+Q, @_jmp_X+Q, @xboot_1+Q

              ' drop refill _jmp _xboot1 ( Refill and loop again )
{1be2 378} xboot_2       word      @drop_X+Q, @refill_X+Q, @_lit_X+Q, 13, @emit_X+Q, @_jmp_X+Q, @xboot_1+Q, 0
{1bf4 37d}               long

{1bf4 37d} switch_L      word      @xboot_L+Q
{1bf6 37d}               byte      FLAG_DEF, 6, "switch"
{1c00 380}               long
{1c00 380} switch_X      word      execlistfunc, 0
{1c04 381}               word      @_lit_X+Q, @getchar_X+Q, @_lit_X+Q, @key_B+Q, @store_X+Q, 0
{1c10 384}               long

{1c10 384} _last         long

{1c10 384} _loop_L       word      @switch_L+Q
{1c12 384}               byte      FLAG_CORE, 5, "_loop"
{1c1c 387}               long
{1c1c 387} _loop_X       word      _loopfunc, 0

{1c20 388} _here         long

'*******************************************************************************
' The Forth source files follow below.  They will be compiled into the
' dictionary, which will over-write the source data.  Some padding space is
' included to ensure that we don't over-write the source data before it is
' compiled.
'*******************************************************************************
{1c20 388}               long      0[100]
{1db0 3ec} infile        file      "init.fth"
{4932 ecc}               file      "comus.fth"
{4c26 f89}               file      "see.fth"
{5061 1098}               file      "p2words.fth"
              'file      "sd.fth"
              'file      "sdutils.fth"
              'file      "linux.fth"
              'file      "cd.fth"
              'file      "ted.fth"
              'file      "bufser.fth"
              'file      "i2c.fth"
              'file      "fds.fth"
              'file      "time.fth"
              'file      "toggle.fth"
{5425 1189}               file      "primes.fth"
              'file      "chess.fth"

'*******************************************************************************
' Enable serial output, print version string and switch to serial input
'*******************************************************************************
{5911 12c4}               byte      13
{5912 12c4}               byte      " 1 verbose !"
{591e 12c7}               byte      " pfthversion type cr"
{5932 12cc}               byte      " switch", 13
              
