#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <ctype.h>
#include "strsubs.h"
#include "symsubs.h"

// Next expected token
#define EXPECT_ANYTHING 0
#define EXPECT_IF_OR_OP 1
#define EXPECT_OP       2
#define EXPECT_DST      3
#define EXPECT_DST1     4
#define EXPECT_DST2     5
#define EXPECT_SRC      6
#define EXPECT_DONE     7
#define EXPECT_WX       8
#define EXPECT_CALL     9
#define EXPECT_DST2P    10
#define EXPECT_DST1P    11
#define EXPECT_OPREP    12
#define EXPECT_UNKNOWN  13
#define EXPECT_OP2A     14
#define EXPECT_OP2B     15
#define EXPECT_OP2C     16
#define EXPECT_OP1A     17
#define EXPECT_OP1B     18
#define EXPECT_OP1C     19
#define EXPECT_OP2D     20
#define EXPECT_OP2E     21
#define EXPECT_RSRC     22

// Printing modes
#define PRINT_NOCODE    0
#define PRINT_CODE      1
#define PRINT_NONE      2

// Spin section types
#define MODE_DAT 0
#define MODE_CON 1
#define MODE_PUB 2
#define MODE_PRI 3
#define MODE_VAR 4

// Keywords and delimiters
char *SectionKeywords[] = {"dat", "con", "pub", "pri", "var", 0};

FILE *infile;
FILE *outfile;

SymbolT SymbolTable[MAX_SYMBOLS];

int numsym = 0;
int cog_addr = 0;
int hub_addr = 0;
int propmode = 1;
int printlevel = 3;
int datcount = 0;
int unicode = 0;
int last_size = 1;
int num_methods = 0;

// This routine handles the P2 pointer syntax
int EncodePointerField(int value, int *pindex, char **tokens, int num)
{
    int i = *pindex;
    int index, retval;
    SymbolT *s;
    int negate = 0;

    retval = 0x100 | ((value & 1) << 7);
    if (i >= num - 1 || strcmp(tokens[i+1], "[")) return retval;

    if (++i > num - 1)
    {
	*pindex = i;
	return -1;
    }

    i++;
    // Check for pre increment or decrement or plus or minus
    if (strcmp(tokens[i], "++") == 0)
    {
	retval |= 0x040;
	i++;
    }
    else if (strcmp(tokens[i], "--") == 0)
    {
	negate = 1;
	retval |= 0x040;
	i++;
    }
    else if (strcmp(tokens[i], "+") == 0)
    {
	i++;
    }
    else if (strcmp(tokens[i], "-") == 0)
    {
	negate = 1;
	i++;
    }

    if (i >= num)
    {
	*pindex = i;
	return -1;
    }
    s = GetSymbolPointer(tokens[i]);
    if (!s)
    {
	*pindex = i;
	return -1;
    }
    i++;
    value = s->value;
    if (negate) value = -value;
    value = Evaluate(value, &i, tokens, num);

    if (i >= num)
    {
	*pindex = i;
	return -1;
    }

    // Check for post increment or decrement
    if (strcmp(tokens[i], "++") == 0)
    {
	retval |= 0x060;
	i++;
    }
    else if (strcmp(tokens[i], "--") == 0)
    {
	value = -value;
	retval |= 0x060;
	i++;
    }

    if (value < -16 || value > 15)
    {
        fprintf(outfile, "Warning: pointer index %d is out of bounds\n", value);
    }

    retval |= (value & 31);

    if (i >= num)
    {
        *pindex = i;
	return -1;
    }

    if (strcmp(tokens[i], "]"))
    {
        fprintf(outfile, "Expected \"]\"\n");
        *pindex = i;
	return -1;
    }

    *pindex = i;
    return retval;
}

// This routine processes the source or destination field
// P2 pointer expressions are allowed if type is >= 2
// Immediate values using a # are allowed if type is >= 1
// The value is returned if successful, and bit at 0x200 is set if immediate
// A minus one is return if unsuccessful
int EncodeAddressField(SymbolT *s, int *pindex, char **tokens, int num, int type)
{
    int retval = 0;
    int i = *pindex;
    int index, value, errnum;

    if (StrCompNoCase(s->name, "ptra") ||
        StrCompNoCase(s->name, "ptrb"))
    {
	if (type >= 2)
	{
            value = EncodePointerField(s->value, &i, tokens, num);
	    if (value >= 0) value |= 0x200;
	}
	else
	{
	    fprintf(outfile, "ERROR: PTR not allowed\n");
	    value = -1;
	}
	*pindex = i;
	return value;
    }

    if (strcmp(s->name, "#") == 0)
    {
	if (type < 1)
	{
	    *pindex = i;
	    return -1;
	}
	retval = 0x200;
	if (++i >= num)
	{
	    *pindex = i;
	    return -1;
	}
        errnum = EvaluateExpression(12, &i, tokens, num, &value);
    }
    else
    {
        errnum = EvaluateExpression(12, &i, tokens, num, &value);
    }

    if (errnum)
    {
        *pindex = i;
        return -1;
    }
    retval |= (value & 511);
    *pindex = i;
    return retval;
}

// This routine returns the next expected state based on the op_type
int GetExpect(int op_type, int i, int num)
{
    int expect;

    if (op_type == TYPE_OP0)
    {
	if (i != num - 1)
            expect = EXPECT_WX;
	else
            expect = EXPECT_DONE;
    }
    else if (op_type == TYPE_OP1S)
        expect = EXPECT_SRC;
    else if (op_type == TYPE_OP1D)
        expect = EXPECT_DST1;
    else if (op_type == TYPE_OP1DN)
        expect = EXPECT_DST2;
    else if (op_type == TYPE_OP2)
        expect = EXPECT_DST;
    else if (op_type == TYPE_OP2P)
        expect = EXPECT_DST2P;
    else if (op_type == TYPE_OPREP)
        expect = EXPECT_OPREP;
    else if (op_type == TYPE_OP1DP)
        expect = EXPECT_DST1P;
    else if (op_type == TYPE_CALL)
        expect = EXPECT_CALL;
    else if (op_type == TYPE_OP2A)
        expect = EXPECT_OP2A;
    else if (op_type == TYPE_OP2B)
        expect = EXPECT_OP2B;
    else if (op_type == TYPE_OP2C)
        expect = EXPECT_OP2C;
    else if (op_type == TYPE_OP2D)
        expect = EXPECT_OP2D;
    else if (op_type == TYPE_OP2E)
        expect = EXPECT_OP2E;
    else if (op_type == TYPE_OP1A)
        expect = EXPECT_OP1A;
    else if (op_type == TYPE_OP1B)
        expect = EXPECT_OP1B;
    else if (op_type == TYPE_OP1C)
        expect = EXPECT_OP1C;
    else
        expect = EXPECT_UNKNOWN;

    return expect;
}

int CountCommas(int i, char **tokens, int num)
{
    int count = 0;

    while (i < num)
    {
        if (!strcmp(tokens[i++], ",")) count++;
    }

    return count;
}

int CountExtras(int i, char **tokens, int num)
{
    int count = 0;

    while (i < num)
    {
        if (!strcmp(tokens[i++], "[")) count += atol(tokens[i]) - 1;
    }

    return count;
}

int CountStrings(int i, char **tokens, int num)
{
    int count = 0;

    for (;i < num; i++)
    {
        if (tokens[i][0] == '"') count += strlen(tokens[i]) - 3;
    }

    return count;
}

int CountElements(int i, char **tokens, int num)
{
    int num_elem = 1;

    if (i + 1 >= num) return 0;    
    num_elem += CountCommas(i, tokens, num);
    num_elem += CountExtras(i, tokens, num);
    num_elem += CountStrings(i, tokens, num);
    return num_elem;
}

void PrintIt(int printflag, int labelflag, char *labelptr, int printlevel, int hub_addr, int cog_addr, int data_size, char *buffer2, int opcode, int num);

//int PrintMore(int printlevel, int hub_addr, int cog_addr, int i, int num, int more_flag, char **tokens);

// Parse a line in the DAT section
void ParseDat(int pass, char *buffer2, char **tokens, int num)
{
    int i;
    SymbolT *s;
    int index;
    int value;

    int rflag = 0;
    int opindex = 0;
    //int more_flag = 0;
    int data_size = 4;
    int cog_incr = 0;
    int hub_incr = 0;
    int labelflag = 0;
    int printflag = PRINT_CODE;
    int opcode = 0x003c0000; // Initialize to if_always
    int expect = EXPECT_ANYTHING;

    // Loop over all the tokens in this line
    for (i = 0; i < num; i++)
    {
        // Find the token is in the symbol table
        index = FindSymbol(tokens[i]);

        // Check for an undefined symbol or local label
        if (index < 0 ||
            (pass == 1 && i == 0 && tokens[i][0] == ':'))
        {
            // If first token add the symbol as an address label
            if (expect == EXPECT_ANYTHING)
	    {
                AddSymbol2(tokens[i], cog_addr >> 2, hub_addr, TYPE_ADDR);
                //fprintf(outfile, "AddSymbol2(%s, %x, %x, %x)\n", tokens[i], cog_addr >> 2, hub_addr, TYPE_ADDR);
	        if (i == num - 1)
	        {
		    printflag = PRINT_NOCODE;
		    expect = EXPECT_DONE;
		    break;
	        }
	        expect = EXPECT_IF_OR_OP;
	        continue;
	    }

            // If pass 1 replace symbol with zero since it may not be defined yet
	    else if (pass == 1)
	    {
	        index = 0;
	        SymbolTable[0].value = 0;
	    }

            // Otherwise, this is an undefined symbol in pass 2
            else
	    {
                fprintf(outfile, "1Unexpected symbol %s\n", tokens[i]);
	        printflag = PRINT_NOCODE;
	        break;
	    }
        }

        // Get a pointer to the symbol in the table
        s = &SymbolTable[index];
        //if (printlevel) printf("%s %x\n", s->name, s->value);

        // Check if a label and process it.  If not a label change
        // expect to EXPECT_IF_OR_OP and continue processing.
        if (expect == EXPECT_ANYTHING)
        {
            // Check if this is an address label
	    if (s->type == TYPE_ADDR)
	    {
	        if (tokens[i][0] != ':')
                {
                    if (num > 1 && last_size != 4)
                        labelflag = 4;
                    else
	                labelflag = 3;
                }
	        if (pass == 2)
	        {
		    if (tokens[i][0] != ':') PurgeLocalLabels(index);
		    if (i == num - 1)
		    {
		        expect = EXPECT_DONE;
		        printflag = PRINT_NOCODE;
		        break;
		    }
		    continue;
	        }

                // If this is pass 1 then this label is already defined
	        printf("label %s is already defined\n", tokens[i]);
	        break;
	    }
	    expect = EXPECT_IF_OR_OP;
        }

        // Handle a conditional, opcode or psuedo-op
        if (expect == EXPECT_IF_OR_OP)
        {
            // Check for conditional
            if (s->type == TYPE_IF)
	    {
	        opcode = s->value;
	        expect = EXPECT_OP;
                cog_addr = (cog_addr + 3) & ~3;
                hub_addr = (hub_addr + 3) & ~3;
	        cog_incr = 4;
	        hub_incr = 4;
                continue;
	    }

            opindex = index;

            // Check for opcode or psuedo-op
            cog_incr = 4;
            hub_incr = 4;
            expect = GetExpect(s->type, i, num);
            if (expect != EXPECT_UNKNOWN)
            {
                // Check for NULL value and type, which is a NOP
                if (s->value || s->type)
                    opcode |= s->value;
                else
                    opcode = 0;
                cog_addr = (cog_addr + 3) & ~3;
                hub_addr = (hub_addr + 3) & ~3;
            }
            else if (s->type == TYPE_ORG)
            {
	        printflag = PRINT_NOCODE;
                hub_addr = (hub_addr + 3) & ~3;
	        if (i == num - 1)
	            cog_addr = 0;
	        else
	        {
                    i++;
                    if (EvaluateExpression(12, &i, tokens, num, &cog_addr))
                        break;
                    cog_addr <<= 2;
                }
	        cog_incr = 0;
	        hub_incr = 0;
	        break;
            }
            else if (s->type == TYPE_RES)
            {
                hub_incr = 0;
                labelflag = 0;
	        printflag = PRINT_NOCODE;
	        if (i == num - 1)
	            cog_incr = 4;
	        else
	        {
                    i++;
                    if (EvaluateExpression(12, &i, tokens, num, &cog_incr))
                        break;
                    cog_incr <<= 2;
	        }
                cog_addr = (cog_addr + 3) & ~3;
                hub_addr = (hub_addr + 3) & ~3;
	        break;
            }
            else if (s->type == TYPE_FIT)
            {
                labelflag = 0;
	        printflag = PRINT_NOCODE;
                cog_incr = 0;
                hub_incr = 0;
	        break;
            }
            else if (s->type == TYPE_LONG)
            {
                cog_addr = (cog_addr + 3) & ~3;
                hub_addr = (hub_addr + 3) & ~3;
                cog_incr = hub_incr = 4 * CountElements(i, tokens, num);
                if (pass == 2)
                    fprintf(outfile, "{%4.4x %3.3x} %s\n", hub_addr, cog_addr >> 2, buffer2);
                cog_addr += cog_incr;
                hub_addr += hub_incr;
                return;
            }
            else if (s->type == TYPE_WORD)
            {
                cog_addr = (cog_addr + 1) & ~1;
                hub_addr = (hub_addr + 1) & ~1;
                cog_incr = hub_incr = 2 * CountElements(i, tokens, num);
                if (pass == 2)
                    fprintf(outfile, "{%4.4x %3.3x} %s\n", hub_addr, cog_addr >> 2, buffer2);
                cog_addr += cog_incr;
                hub_addr += hub_incr;
                return;
            }
            else if (s->type == TYPE_BYTE)
            {
                cog_incr = hub_incr = CountElements(i, tokens, num);
                if (pass == 2)
                    fprintf(outfile, "{%4.4x %3.3x} %s\n", hub_addr, cog_addr >> 2, buffer2);
                cog_addr += cog_incr;
                hub_addr += hub_incr;
                return;
            }
            else if (s->type == TYPE_FILE)
            {
                char filename[100];
                int len, filesize = 0;
                FILE *tempfile;
                strcpy(filename, &tokens[i+1][1]);
                len = strlen(filename);
                if (len) filename[len-1] = 0;
                tempfile = fopen(filename, "r");
                if (!tempfile)
                {
                    fprintf(outfile, "ERROR: Couldn't open %s\n", filename);
                }
                else
                {
                    fseek(tempfile, 0, SEEK_END);
                    filesize = ftell(tempfile);
                    fclose(tempfile);
                }
                cog_incr = hub_incr = filesize;
                if (pass == 2)
                    fprintf(outfile, "{%4.4x %3.3x} %s\n", hub_addr, cog_addr >> 2, buffer2);
                cog_addr += cog_incr;
                hub_addr += hub_incr;
                return;
            }
            else
	    {
                fprintf(outfile, "4Unexpected symbol %s\n", tokens[i]);
	        break;
	    }
	    if (pass == 1) break;
        }

        // Handle an opcode
        else if (expect == EXPECT_OP)
        {
	    cog_incr = 4;
	    hub_incr = 4;
            expect = GetExpect(s->type, i, num);
            if (expect != EXPECT_UNKNOWN)
            {
	        // Check for NULL value, which is a NOP
                if (s->value || s->type)
	            opcode |= s->value;
	        else
	            opcode = 0;
            }
            else
	    {
                fprintf(outfile, "5Unexpected symbol %s\n", tokens[i]);
	        break;
	    }
            opindex = index;
	    if (pass == 1) break;
        }

        // Handle destination for OP2 instruction, such as add d,s
        else if (expect == EXPECT_DST)
        {
            if (EvaluateExpression(12, &i, tokens, num, &value)) break;
	    opcode |= (value & 511) << 9;
	    i++;
	    if (i == num || strcmp(tokens[i], ","))
	    {
	        fprintf(outfile, "Expected a comma\n");
	        break;
	    }
	    expect = EXPECT_SRC;
        }

        // Handle destination for OP1D instruction, such as locknew d
        else if (expect == EXPECT_DST1)
        {
            if (EvaluateExpression(12, &i, tokens, num, &value)) break;
	    opcode |= (value & 511) << 9;
	    if (i != num - 1)
	        expect = EXPECT_WX;
	    else
	        expect = EXPECT_DONE;
        }

        // Handle destination for OP1DN instruction, such as setptr d/#n
        else if (expect == EXPECT_DST2)
        {
	    if (strcmp(s->name, "#"))
	    {
	        value = Evaluate(s->value, &i, tokens, num);
	        opcode |= (value & 511) << 9;
	        if (i != num - 1)
	            expect = EXPECT_WX;
	        else
	            expect = EXPECT_DONE;
	    }
	    else
	    {
	        opcode |= 1 << 23;
	    }
        }

        // Handle destination for OP2A instruction, such as djnz d,s/@
        else if (expect == EXPECT_OP2A)
        {
            if (EvaluateExpression(12, &i, tokens, num, &value)) break;
	    opcode |= (value & 511) << 9;
	    i++;
	    if (i == num || strcmp(tokens[i], ","))
	    {
	        fprintf(outfile, "Expected a comma\n");
	        break;
	    }
	    expect = EXPECT_RSRC;
        }

        // Handle destination for OP2B instruction, such as rep d/#,s/#
        else if (expect == EXPECT_OP2B)
        {
	    if (strcmp(s->name, "#"))
	    {
                if (EvaluateExpression(12, &i, tokens, num, &value)) break;
	        opcode |= (value & 511) << 9;
	        i++;
	        if (i == num || strcmp(tokens[i], ","))
	        {
	            fprintf(outfile, "Expected a comma\n");
	            break;
	        }
	        expect = EXPECT_SRC;
	    }
	    else
	    {
	        opcode |= 1 << 23;
	    }
        }

        // Handle destination for OP2C instruction, such as wrbyte d/#,s/ptr
        else if (expect == EXPECT_OP2C)
        {
	    if (strcmp(s->name, "#"))
	    {
                if (EvaluateExpression(12, &i, tokens, num, &value)) break;
	        opcode |= (value & 511) << 9;
	        i++;
	        if (i == num || strcmp(tokens[i], ","))
	        {
	            fprintf(outfile, "Expected a comma\n");
	            break;
	        }
	        expect = EXPECT_SRC;
	    }
	    else
	    {
	        opcode |= 1 << 23;
	    }
        }

        // Handle destination for OP2D instruction, such as jp d/#,s/@
        else if (expect == EXPECT_OP2D)
        {
	    if (strcmp(s->name, "#"))
	    {
                if (EvaluateExpression(12, &i, tokens, num, &value)) break;
	        opcode |= (value & 511) << 9;
	        i++;
	        if (i == num || strcmp(tokens[i], ","))
	        {
	            fprintf(outfile, "Expected a comma\n");
	            break;
	        }
	        expect = EXPECT_RSRC;
	    }
	    else
	    {
	        opcode |= 1 << 23;
	    }
        }

        // Handle destination for OP2E instruction, such as calld reg,#absolute/@relative
        else if (expect == EXPECT_OP2E)
        {
	    value = Evaluate(s->value, &i, tokens, num);
	    opcode |= ((value - 0x1f6) & 2) << 25;
	    if (strcmp(tokens[i], ","))
	    {
	        fprintf(outfile, "ERROR: Expected a comma -- %s\n", tokens[i]);
	        break;
	    }
	    if (++i >= num) break;
            expect = EXPECT_OP1A;
        }

        // Handle destination for OP1A instruction, such as jmp #absolute/@relative
        else if (expect == EXPECT_OP1A)
        {
            //printf("EXPECT_OP1A: opcode = %s\n", SymbolTable[opindex].name);
	    if (strcmp(s->name, "@") && strcmp(s->name, "#"))
            {
                if (!strcmp(SymbolTable[opindex].name, "jmp"))
                {
	            fprintf(outfile, "ERROR: Expected a '#' or '@' after jmp instruction\n");
	            break;
                }
                s = &SymbolTable[opindex+1];
                opcode = s->value | (opcode & 0x003c0000);
                expect = EXPECT_DST1;
                i--;
                continue;
            }
	    if (!strcmp(s->name, "@"))
            {
                rflag = 1;
	        opcode |= 1 << 24;
            }
            i++;
	    if (EvaluateExpression(12, &i, tokens, num, &value)) break;
            //if (rflag) printf("value = %x - (%x >> 2) = %x\n", value, cog_addr, value - (cog_addr >> 2));
            if (rflag) value -= cog_addr >> 2;
	    opcode |= (value & 0x3ffff);
            opcode |= (value << 4) & 0xc00000;
	    if (i != num - 1)
	        expect = EXPECT_WX;
	    else
	        expect = EXPECT_DONE;
        }

        // Handle destination for OP1B instruction, such as augs #23bits
        else if (expect == EXPECT_OP1B)
        {
	    if (strcmp(s->name, "#"))
	    {
	        value = Evaluate(s->value, &i, tokens, num);
	        opcode |= (value & 0x3ffff);
                opcode |= (value << 4) & 0x7c00000;
	        expect = EXPECT_DONE;
	    }
        }

        // Handle destination for OP1C instruction, such as cogstop d/#
        else if (expect == EXPECT_OP1C)
        {
	    if (strcmp(s->name, "#"))
	    {
	        value = Evaluate(s->value, &i, tokens, num);
	        opcode |= (value & 511) << 9;
	        if (i != num - 1)
	            expect = EXPECT_WX;
	        else
	            expect = EXPECT_DONE;
	    }
	    else
	    {
	        opcode |= 1 << 22;
	    }
        }

        // Handle destination for OP1DP instruction, such as rdquad d/#n/ptr
        else if (expect == EXPECT_DST1P)
        {
	    value = EncodeAddressField(s, &i, tokens, num, 2);
	    if (value < 0) break;
	    if (value & 0x200) opcode |= 0x00800000;
	    opcode |= (value & 0x1ff) << 9;
	    if (i < num - 1)
	        expect = EXPECT_WX;
	    else
	        expect = EXPECT_DONE;
        }

        // Handle destination for OP2P or OPREP instruction, such as
        // rdbyte d,s/#n/ptr or rep d/#n,i
        else if (expect == EXPECT_DST2P || expect == EXPECT_OPREP)
        {
	    if (expect == EXPECT_OPREP)
	    {
	        value = EncodeAddressField(s, &i, tokens, num, 1);
	        if (value < 0) break;
	        if (value & 0x200) opcode |= 0x00800000;
	    }
	    else
	    {
	        value = EncodeAddressField(s, &i, tokens, num, 0);
	        if (value < 0 || value > 511) break;
	    }
	    opcode |= (value & 0x1ff) << 9;
	    if (++i >= num) break;
	    if (strcmp(tokens[i], ","))
	    {
	        fprintf(outfile, "ERROR: Expected a comma -- %s\n", tokens[i]);
	        break;
	    }
	    if (++i >= num) break;
	    s = GetSymbolPointer(tokens[i]);
	    if (!s) break;
	    value = EncodeAddressField(s, &i, tokens, num, 2);
	    if (value < 0) break;
	    if (value & 0x200) opcode |= 0x00400000;
	    opcode |= (value & 0x1ff);
	    if (i < num - 1)
	        expect = EXPECT_WX;
	    else
	        expect = EXPECT_DONE;
        }

        // Handle source for OP1S or OP2 instruction,
        // such as jmp s or add d,s/#n
        else if (expect == EXPECT_SRC)
        {
	    if (strcmp(s->name, "#"))
	    {
	        if (EvaluateExpression(12, &i, tokens, num, &value)) break;
	        opcode |= (value & 511);
	        if (i != num - 1)
	            expect = EXPECT_WX;
	        else
	            expect = EXPECT_DONE;
	    }
	    else
	    {
	        opcode |= s->value;
	    }
        }

        // Handle source for instructions such as djnz d,s/@
        else if (expect == EXPECT_RSRC)
        {
	    if (strcmp(s->name, "@"))
	    {
	        if (EvaluateExpression(12, &i, tokens, num, &value)) break;
                //if (rflag) printf("value = %x - (%x >> 2) = %x\n", value, cog_addr, value - (cog_addr >> 2));
                if (rflag) value -= cog_addr >> 2;
	        opcode |= (value & 511);
	        if (i != num - 1)
	            expect = EXPECT_WX;
	        else
	            expect = EXPECT_DONE;
	    }
	    else
	    {
                rflag = 1;
	        opcode |= 1 << 22;
	    }
        }

        // Handle source and destination for call instruction
        else if (expect == EXPECT_CALL)
        {
	    if (strcmp(s->name, "#"))
	    {
	        char tempbuf[200];

	        opcode |= (s->value & 511);
	        strcpy(tempbuf, tokens[i]);
	        strcat(tempbuf, "_ret");
                s = GetSymbolPointer(tempbuf);
	        if (!s) break;
	        opcode |= (s->value & 511) << 9;
	        if (i != num - 1)
	            expect = EXPECT_WX;
	        else
	            expect = EXPECT_DONE;
	    }
	    else
	    {
	        opcode |= s->value;
	    }
        }

        // Handle conditional flags, such as wc or nr
        else if (expect == EXPECT_WX)
        {
	    if (s->type != TYPE_WX && s->type != TYPE_NX)
	    {
                fprintf(outfile, "6Unexpected symbol %s\n", tokens[i]);
	        break;
	    }
	    if (s->type == TYPE_WX)
	        opcode |= s->value;
	    else
	        opcode &= ~s->value;
	    if (++i == num)
	    {
	        expect = EXPECT_DONE;
	        break;
	    }
	    if (strcmp(tokens[i], ","))
	    {
                fprintf(outfile, "7Unexpected symbol %s\n", tokens[i]);
	        break;
	    }
        }

        // Check if there are too many tokens
        else if (expect == EXPECT_DONE)
        {
            fprintf(outfile, "8Unexpected symbol %s\n", tokens[i]);
	    break;
        }
    }

    // Check if there were not enough tokens
    if (expect != EXPECT_DONE && i == num)
    {
        fprintf(outfile, "Unexpected EOL\n");
    }

    if (pass == 2)
    {
        PrintIt(printflag, labelflag, tokens[0], printlevel, hub_addr, cog_addr, data_size, buffer2, opcode, num);
    }
 
    // Update the address
    cog_addr += cog_incr;
    hub_addr += hub_incr;
    if (data_size) last_size = data_size;

#if 0
    if (more_flag)
    {
        cog_incr = PrintMore(printlevel, hub_addr, cog_addr, i, num, more_flag, tokens);
        // Update the address
        cog_addr += cog_incr;
        hub_addr += cog_incr;
        last_size = more_flag;
    }
#endif
}

#if 0
int PrintMore(int printlevel, int hub_addr, int cog_addr, int i, int num, int more_flag, char **tokens)
{
    char *labeltypes[] = {" byte", " word", "", " long"};
    int count = 0;
    int value;
    SymbolT *s;
    int incr = 0;

    while (i < num)
    {
        if (count++ == 0)
        {
            if (printlevel >= 3)
                fprintf(outfile, "{%4.4x %3.3x} ", hub_addr, cog_addr >> 2);
            fprintf(outfile, " %s", labeltypes[more_flag-1]);
        }
        if (strcmp(tokens[i], ","))
        {
            if (count) fprintf(outfile, "\n");
            fprintf(outfile, "\nExpected a comma\n");
            return;
        }
        if (++i >= num)
        {
            fprintf(outfile, "\nUnexpected EOL\n");
            return;
        }
        s = GetSymbolPointer(tokens[i]);
        value = Evaluate(s->value, &i, tokens, num);
        fprintf(outfile, " $%8.8x", value);
        hub_addr += more_flag;
        cog_addr += more_flag;
        incr += more_flag;
        if (++i < num && count < 6) fprintf(outfile, ",");
        else
        {
            count = 0;
            fprintf(outfile, "\n");
        }
    }
    return incr;
}
#endif

void PrintIt(int printflag, int labelflag, char *labelptr, int printlevel, int hub_addr, int cog_addr, int data_size, char *buffer2, int opcode, int num)
{
    char *labeltypes[] = {" byte", " word", "", " long"};

    if (printlevel == 0) return;

    if (printlevel >= 1 && datcount)
    {
        datcount = 0;
        fprintf(outfile, "\n");
    }

    if (printlevel >= 1 && labelflag)
    {
        if (printlevel >= 3)
            fprintf(outfile, "{%4.4x %3.3x} ", hub_addr, cog_addr >> 2);
        fprintf(outfile, "%s%s\n", labelptr, labeltypes[labelflag-1]);
    }

    if (labelflag && num == 1) return;

    if (printlevel >= 3)
        fprintf(outfile, "{%4.4x %3.3x} ", hub_addr, cog_addr >> 2);

    if (printflag == PRINT_NOCODE)
    {
        fprintf(outfile, "%s\n", buffer2);
        return;
    }

    if (data_size == 1)
        fprintf(outfile, "  byte $%2.2x       ", opcode);
    else if (data_size == 2)
        fprintf(outfile, "  word $%4.4x     ", opcode);
    else
        fprintf(outfile, "  long $%8.8x ", opcode);

    if (printlevel >= 2)
        fprintf(outfile, " '%s\n", buffer2);
    else
        fprintf(outfile, "\n");
}

int CheckComment(char *buffer, int *pflag)
{
    int len = strlen(buffer);

    if (len == 0)
    {
        return *pflag;
    }
    else if (*pflag)
    {
        *pflag = (buffer[len-1] != '}');
        return 1;
    }
    else
    {
        char *ptr = SkipChars(buffer, " \t");
        *pflag = (*ptr == '{');
        return *pflag;
    }
}

int ProcessConstantLine(int currval, char **tokens, int num)
{
    int j;
    int i= 0;
    SymbolT *s;
    int commaflag = 0;

    if (num < 1) return currval;

    if (StrCompNoCase(tokens[0], "con"))
    {
        i++;
        currval = 0;
    }

    for (;i < num; i++)
    {
        if (commaflag)
        {
            commaflag = 0;
            if (strcmp(tokens[i], ","))
            {
                fprintf(outfile, "Expected a comma\n");
                break;
            }
        }
        else
        {
            commaflag = 1;
            if (!strcmp(tokens[i], "#"))
            {
                if (i >= num - 1)
                {
                    fprintf(outfile, "Expected a constant value\n");
                    break;
                }
                s = GetSymbolPointer(tokens[++i]);
                if (s) currval = Evaluate(s->value, &i, tokens, num);
            }
            else if (i < num - 1 && !strcmp(tokens[i+1], "="))
            {
                if (i >= num - 2)
                {
                    fprintf(outfile, "Expected a constant value\n");
                    break;
                }
                j = i;
                i += 2;
                s = GetSymbolPointer(tokens[i]);
                if (s) currval = Evaluate(s->value, &i, tokens, num);
                AddSymbol(tokens[j], currval++, TYPE_CON);
            }
            else
            {
                AddSymbol(tokens[i], currval++, TYPE_CON);
            }
        }
    }

    return currval;
}

void CountMethods(void)
{
    int i, num;
    char *tokens[100];
    char buffer[300];
    char buffer2[300];
    int mode = MODE_CON;
    int commentflag = 0;

    while (ReadString(buffer, 300, infile, unicode))
    {
        if (CheckComment(buffer, &commentflag)) continue;
	num = Tokenize(buffer, tokens, 100, buffer2);
	if (num > 0) i = SearchList(SectionKeywords, tokens[0]);
	else i = -1;
	if (i >= 0 || mode != MODE_DAT || num == 0)
	{
            if (i == MODE_PUB || i == MODE_PRI) num_methods++;
	    if (i >= 0) mode = i;
        }
    }
    //printf("num_methods = %d\n", num_methods);
}

// Parse a Spin file and call ParseDat for lines in the DAT section
void Parse(int pass)
{
    int i, num;
    char *tokens[100];
    char buffer[300];
    char buffer2[300];
    SymbolT *s;
    int expect, index;
    int value;
    int incr;
    int mode = MODE_CON;
    int commentflag = 0;
    int currval = 0;

    cog_addr = 0;
    hub_addr = (num_methods + 1) * 4 + 16;
    datcount = 0;
    last_size = 1;

    while (ReadString(buffer2, 300, infile, unicode))
    {
        if (CheckComment(buffer2, &commentflag))
        {
	    if (pass == 2) fprintf(outfile, "%s\n", buffer2);
            continue;
        }
	//strcpy(buffer2, buffer);
	num = Tokenize(buffer2, tokens, 100, buffer);
	if (num > 0) i = SearchList(SectionKeywords, tokens[0]);
	else i = -1;
	if (i >= 0 || mode != MODE_DAT || num == 0)
	{
	    if (i >= 0) mode = i;
	    if (pass == 2) fprintf(outfile, "%s\n", buffer2);
	    else if (mode == MODE_CON)
                currval = ProcessConstantLine(currval, tokens, num);
	    continue;
        }
        ParseDat(pass, buffer2, tokens, num);
    }
    if (datcount && printlevel == 0) fprintf(outfile, "\n");
}

void usage(void)
{
    printf("usage: qasm [options] file\n");
    printf("     -1           Prop 1 mode (default)\n");
    printf("     -2           Prop 2 mode\n");
    printf("     -o filename  Set output file name (default qasmout.spin)\n");
    printf("     -p#          Set print level\n");
    exit(1);
}

int main(int argc, char **argv)
{
    int i;
    char *ptr;
    char *infname = 0;
    char *outfname = "qasmout.spin";

    for (i = 1; i < argc; i++)
    {
	if (argv[i][0] == '-')
	{
	    if (strcmp(argv[i], "-1") == 0)
	        propmode = 1;
	    else if (strcmp(argv[i], "-2") == 0)
	        propmode = 2;
	    else if (strcmp(argv[i], "-o") == 0 && i < argc - 1)
                outfname = argv[++i];
	    else if (strncmp(argv[i], "-p", 2) == 0)
	        printlevel = argv[i][2] - '0';
	    else
	        usage();
	}
	else
           infname = argv[i];
    }

    if (!infname) usage();

    infile = FileOpen(infname, "r");
    unicode = CheckForUnicode(infile);
    outfile = FileOpen(outfname, "w");

    ReadSymbolTable();
    CountMethods();
    Parse(1);
    Parse(2);

    return 0;
}
