#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <ctype.h>
#include "strsubs.h"
#include "symsubs.h"

extern int hub_addr;
extern int propmode;
extern int numsym;
extern FILE *outfile;

extern SymbolT SymbolTable[MAX_SYMBOLS];

void ReadSymbolTable(void)
{
    SymbolT *s;
    FILE *symfile;
    char buffer[200];

    if (propmode == 1)
        symfile = FileOpen("pasmsym1.txt", "r");
    else
        symfile = FileOpen("pasmsym2.txt", "r");

    s = &SymbolTable[0];

    s->name[0] = 0;
    s->value = 0;
    s->type = TYPE_CON;
    numsym = 1;
    s = &SymbolTable[1];

    while (fgets(buffer, 200, symfile))
    {
	RemoveCRLF(buffer);
	if (buffer[0] == 0 || buffer[0] == '\'') continue;
        sscanf(buffer, "%x %d %s", &s->value, &s->type, s->name);
	//printf("ReadSymbolTable: %8.8x %d %s\n", s->value, s->type, s->name);
	s = &SymbolTable[++numsym];
    }
    fclose(symfile);
}

int FindSymbol(char *symbol)
{
    int i;

    if (*symbol == '$')
    {
	if (symbol[1] == 0)
	    SymbolTable[0].value = hub_addr;
	else
	    SymbolTable[0].value = StrToHex(symbol+1);
	    //sscanf(symbol+1, "%x", &SymbolTable[0].value);
	return 0;
    }
    else if (*symbol == '%')
    {
        if (symbol[1] == '%')
	    SymbolTable[0].value = StrToQuad(symbol+2);
        else
	    SymbolTable[0].value = StrToBin(symbol+1);
	//printf("%s = $%x\n", symbol, SymbolTable[0].value);
	return 0;
    }
    else if (*symbol == '"')
    {
	SymbolTable[0].value = symbol[1];
	return 0;
    }
    else if (*symbol == '-' || (*symbol >= '0' && *symbol <= '9'))
    {
        SymbolTable[0].value = StrToDec(symbol);
	//sscanf(symbol, "%d", &SymbolTable[0].value);
	return 0;
    }

    for (i = 0; i < numsym; i++)
    {
	//if (strcmp(symbol, SymbolTable[i].name) == 0) return i;
	if (StrCompNoCase(symbol, SymbolTable[i].name)) return i;
    }
    return -1;
}

void AddSymbol(char *symbol, int value, int type)
{
    if (numsym >= MAX_SYMBOLS)
    {
	printf("Symbol table is full\n");
	exit(1);
    }
    if (strlen(symbol) > 30)
    {
        fprintf(outfile, "Truncating %s to 30 characters\n", symbol);
        symbol[30] = 0;
    }
    strcpy(SymbolTable[numsym].name, symbol);
    SymbolTable[numsym].value = value;
    SymbolTable[numsym].value2 = 0;
    SymbolTable[numsym].type = type;
    numsym++;
}

void AddSymbol2(char *symbol, int value, int value2, int type)
{
    if (numsym >= MAX_SYMBOLS)
    {
	printf("Symbol table is full\n");
	exit(1);
    }
    if (strlen(symbol) > 30)
    {
        fprintf(outfile, "Truncating %s to 30 characters\n", symbol);
        symbol[30] = 0;
    }
    strcpy(SymbolTable[numsym].name, symbol);
    SymbolTable[numsym].value = value;
    SymbolTable[numsym].value2 = value2;
    SymbolTable[numsym].type = type;
    numsym++;
}

void PurgeLocalLabels(int index)
{
    int i;

    for (i = 0; i < index; i++)
    {
	if (SymbolTable[i].name[0] == ':') SymbolTable[i].name[0] = ';';
    }
}

SymbolT *GetSymbolPointer(char *str)
{
    int index = FindSymbol(str);
    if (index < 0)
    {
	fprintf(outfile, "ERROR: %s is undefined\n", str);
	return 0;
    }
    return &SymbolTable[index];
}

int Evaluate(int value, int *pindex, char **tokens, int num)
{
    int index;
    char *ptr;
    SymbolT *s;
    int i = *pindex;
    static char *ops = "+-*/&|<>";

    while (i < num - 2)
    {
	//if (tokens[i+1][1]) break;
	ptr = FindChar(ops, tokens[i+1][0]);
	if (*ptr == 0) break;
	index = FindSymbol(tokens[i+2]);
	if (index < 0)
        {
            printf("Evaluate: Symbol %s is undefined\n", tokens[i+2]);
            break;
        }
	s = &SymbolTable[index];
	if (!strcmp(tokens[i+1], "+")) value += s->value;
	else if (!strcmp(tokens[i+1], "-")) value -= s->value;
	else if (!strcmp(tokens[i+1], "*")) value *= s->value;
	else if (!strcmp(tokens[i+1], "/")) value /= s->value;
	else if (!strcmp(tokens[i+1], "&")) value &= s->value;
	else if (!strcmp(tokens[i+1], "|")) value |= s->value;
	else if (!strcmp(tokens[i+1], ">>")) value >>= s->value;
	else if (!strcmp(tokens[i+1], "<<")) value <<= s->value;
	else break;
	i += 2;
    }

    *pindex = i;
    return value;
}

int EvaluateExpression(int prevprec, int *pindex, char **tokens, int num, int *pval)
{
    int index;
    char *ptr;
    SymbolT *s;
    int i = *pindex;
    static char *oplist[] = {"+", "-", "*", "/", "&", "|", "<<", ">>", 0};
    static int precedence[] = {6, 6, 5, 5, 3, 4, 2, 2};
    int value = 0;
    int currprec;
    int value2;
    int errnum;

    if (i >= num)
    {
        printf("Encountered EOL\n");
        return 1;
    }

    if (!strcmp(tokens[i], "("))
    {
        i++;
        if ((errnum = EvaluateExpression(12, &i, tokens, num, &value)))
            return errnum;
        i++;
        if (i >= num)
        {
            printf("Encountered EOL\n");
            *pindex = i;
            return 1;
        }
        if (strcmp(tokens[i], ")"))
        {
            printf("Expected a closing parenthesis\n");
            *pindex = i;
            return 2;
        }
    }
    else
    {
        int hub_addr_flag = 0;
        if (!strcmp(tokens[i], "@@@"))
        {
            if (++i >= num)
            {
                printf("Encountered EOL\n");
                *pindex = i;
                return 1;
            }
            hub_addr_flag = 1;
        }
        index = FindSymbol(tokens[i]);
        if (index < 0)
        {
            printf("EvaluateExpression: Symbol %s is undefined\n", tokens[i]);
            return 3;
        }
        s = &SymbolTable[index];
        if (hub_addr_flag)
            value = s->value2;
        else
            value = s->value;
    }

    while (i < num - 2)
    {
        index = SearchList(oplist, tokens[i+1]);
        if (index < 0) break;
        currprec = precedence[index];
        if (currprec >= prevprec) break;
        i += 2;
        if ((errnum = EvaluateExpression(currprec, &i, tokens, num, &value2)))
            return errnum;
	if (index == 0)      value += value2;
	else if (index == 1) value -= value2;
	else if (index == 2) value *= value2;
	else if (index == 3) value /= value2;
	else if (index == 4) value &= value2;
	else if (index == 5) value |= value2;
	else if (index == 6) value <<= value2;
	else if (index == 7) value >>= value2;
	else break;
    }

    *pval = value;
    *pindex = i;
    return 0;
}
