/*******************************************************************************
' Author: Dave Hein
' Version 0.21
' Copyright (c) 2010, 2011
' See end of file for terms of use.
'******************************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include "interp.h"
#include "spinsim.h"

#define NUM_ICACHE 1

#define IGNORE_WZ_WC
#define PRINT_RAM_ACCESS

#define REG_CNT  0x1f1
#define REG_PINA 0x1fa
#define REG_PINB 0x1fb
#define REG_OUTA 0x1fc
#define REG_OUTB 0x1fd
#define REG_DIRA 0x1fe
#define REG_DIRB 0x1ff

extern char *hubram;
extern int32_t memsize;
extern char lockstate[8];
extern char lockalloc[8];
extern PasmVarsT PasmVars[8];
extern int32_t pasmspin;
extern int32_t cycleaccurate;
extern int32_t loopcount;
extern int32_t pin_val;
extern SerialT serial_in;
extern SerialT serial_out;
extern int32_t nohubslots;

extern FILE *tracefile;

char *GetOpname2(unsigned int, int *);

static void NotImplemented(int instruction)
{
    int dummy;
    char *opname = GetOpname2(instruction, &dummy);
    printf("\n%s not implemented - %8.8x\n", opname, instruction);
    spinsim_exit(1);
}

static int32_t parity(int32_t val)
{
    val ^= val >> 16;
    val ^= val >> 8;
    val ^= val >> 4;
    val ^= val >> 2;
    val ^= val >> 1;
    return val & 1;
}

static int32_t _abs(int32_t val)
{
    return val < 0 ? -val : val;
}

#if 0
static int32_t CheckWaitPin(PasmVarsT *pasmvars, int32_t instruct, int32_t value1, int32_t value2)
{
    int32_t match, pin_values;
    int32_t portnum = (instruct >> 24) & 3;

    if (portnum == 0)
	pin_values = pin_val;
    else
	pin_values = 0xffffffff;

    if (instruct & 0x04000000)
	match = ((pin_values & value2) != value1); // waitpne
    else
	match = ((pin_values & value2) == value1); // waitpne

    return match;
}
#endif

#if 0
static int wait1[128] = {
    7, 7, 7, 7, 0, 0, 0, 0,    0, 0, 0, 0, 0, 0, 0, 0,
    0, 0, 0, 0, 0, 0, 0, 0,    0, 0, 0, 0, 0, 0, 0, 0,
    0, 0, 0, 0, 0, 0, 0, 0,    0, 0, 0, 0, 0, 0, 0, 0,
    0, 0, 0, 0, 0, 0, 0, 0,    0, 0, 0, 0, 0, 0, 0, 9,
    0, 0, 0, 0, 0, 0, 0, 0,    0, 0, 0, 0, 0, 0, 0, 0,
    0, 0, 0, 0, 0, 0, 0, 0,    0, 0, 0, 0, 0, 0, 0, 0,
    0, 0, 0, 0, 0, 0, 0, 0,    0, 0, 0, 0, 0, 0, 0, 0,
    0, 0, 0, 0, 0, 0, 0, 0,    0, 0, 0, 0, 0, 0, 0, 0};
#endif

#define WAITFLAG_HUB     7
#define WAITFLAG_WAITCNT 9

static int32_t CheckWaitFlag2(PasmVarsT *pasmvars, int instruct, int value1, int value2)
{
    int32_t hubcycles;
    int32_t waitmode = 0;
    //int32_t srcaddr = instruct & 511;
    int32_t waitflag = pasmvars->waitflag;
    int32_t opcode = (instruct >> (32 - 7)) & 127;
    int32_t hubop = 0;

    if (waitflag)
    {
	waitflag--;
#if 0
        if ((instruct & 0xf8000000) == 0xc8000000)
	{
	    if (CheckWaitPin(pasmvars, instruct, value1, value2))
		waitflag = 0;
	}
#endif
	pasmvars->waitflag = waitflag;
	if (waitflag == 0) pasmvars->waitmode = 0;
	return waitflag;
    }

#if 0
    if (opcode != 127)
	waitflag = wait1[opcode];
#if 0
    else if (srcaddr < 0x120)
	waitflag = wait2[srcaddr];
#endif
#else
    if (opcode == 0x54)
        waitflag = WAITFLAG_WAITCNT;
    else if (opcode >= 0x61 && opcode <= 0x66)
        waitflag = WAITFLAG_HUB;
#endif
    else
        waitflag = 0;

    //if (waitflag) printf("waitflag = %d\n", waitflag);

    // if (waitflag) printf("CheckWaitFlag2: %8.8x %d\n", instruct, waitflag);

    if (nohubslots)
	hubcycles = 0;
    else
	hubcycles = (pasmvars->cogid + (value2 >> 2) - loopcount) & 15;

    switch (waitflag)
    {
	case 0: // 1
	break;

	case 1: // 1 - 8
	hubop = 1;
	waitflag = hubcycles;
	waitmode = WAIT_HUB;
	break;

	case 2: // 2
	waitflag = 1;
	waitmode = WAIT_MULT;
	break;

	case 3: // 1 or 3 - 10 rdxxxxc
//printf("\nvalue2 = %8.8x, dcachehubaddr = %8.8x\n", value2&0xffffffe0, pasmvars->dcachehubaddr);
	if ((value2 & 0xfffffe0) == pasmvars->dcachehubaddr)
	    waitflag = 0;
	else
	{
	    hubop = 1;
	    waitmode = WAIT_HUB;
	    waitflag = hubcycles + 2;
	}
	break;

	case 4: // 1 or 1 - 8
	if ((value2 & 0xfffffe0) == pasmvars->dcachehubaddr)
	    waitflag = 0;
	else
	{
	    hubop = 1;
	    waitmode = WAIT_HUB;
	    waitflag = hubcycles;
	}
	break;

	case 5: // 1 - 9 for lockset and lockclr
	hubop = 1;
	waitmode = WAIT_HUB;
	waitflag = hubcycles;
	if (instruct & 0x00800000)
	    waitflag++;
	break;

	case 6: // 2 - 9
	hubop = 1;
	waitmode = WAIT_HUB;
	waitflag = hubcycles + 1;
	break;

	case WAITFLAG_HUB: // 3 - 10 rdxxxx
	hubop = 1;
	waitmode = WAIT_HUB;
	waitflag = hubcycles + 2;
        //printf("WAITFLAG_HUB: waitflag = %x cycles, (value2 >> 2) = %x\n", waitflag, value2 >> 2);
	break;

	case 8: // mac
	waitflag = 0;
	break;

	case WAITFLAG_WAITCNT: // N for waitcnt
	waitmode = WAIT_CNT;
	waitflag = value1 - GetCnt();
	break;

	case 10: // 1 - 8 for wrlong or 1 for frac
	if (instruct & 0x01000000)
	    waitflag = 0;
	else
	{
	    hubop = 1;
	    waitmode = WAIT_HUB;
	    waitflag = hubcycles;
	}
	break;

	case 11: // N for wait
	if (!value1)
	    waitflag = 0;
	else
	{
	    waitmode = WAIT_CNT;
	    waitflag = value1 - 1;
	}
	break;

	case 12: // mulcount
	waitmode = WAIT_MULT;
	waitflag = pasmvars->mulcount;
	break;

	case 13: // 1 - 8 for callax, callbx
	if ((instruct & 0x01800000) == 0x01000000)
	{
	    hubop = 1;
	    waitmode = WAIT_HUB;
	    waitflag = hubcycles;
	}
	else
	    waitflag = 0;
	break;

	case 14: // 1 - 8 for callax, callbx
	hubop = 1;
	waitmode = WAIT_HUB;
	waitflag = hubcycles;
	break;

	case 15: // 2 - 9 for retax, retbx
	hubop = 1;
	waitmode = WAIT_HUB;
	waitflag = hubcycles + 1;
	break;

	case 16: // N for waitpxx
	waitmode = WAIT_PIN;
	waitflag = pasmvars->lastd - GetCnt();
	break;
    }

    pasmvars->waitflag = waitflag;
    pasmvars->waitmode = waitmode;

    return hubop;
}

// Compute the hub RAM address from the pointer instruction - SUPIIIIII
static int32_t GetPointer(PasmVarsT *pasmvars, int32_t ptrinst, int32_t size)
{
    int32_t address = 0; // Set to zero to avoid compiler warning
    int32_t offset = (ptrinst << 26) >> (26 - size);

    switch ((ptrinst >> 6) & 7)
    {
        case 0: // ptra[offset]
        address = (pasmvars->ptra + offset) & 0x3ffff;
        break;

        case 1: // ptra
        address = pasmvars->ptra;
        break;

        case 2: // ptra[++offset]
        address = (pasmvars->ptra + offset) & 0x3ffff;
        //pasmvars->ptra = address;
        pasmvars->ptra0 = address;
        break;

        case 3: // ptra[offset++]
        address = pasmvars->ptra;
        //pasmvars->ptra = (pasmvars->ptra + offset) & 0x3ffff;
        pasmvars->ptra0 = (pasmvars->ptra + offset) & 0x3ffff;
        break;

        case 4: // ptrb[offset]
        address = (pasmvars->ptrb + offset) & 0x3ffff;
        break;

        case 5: // ptrb
        address = pasmvars->ptrb;
        break;

        case 6: // ptrb[++offset]
        address = (pasmvars->ptrb + offset) & 0x3ffff;
        //pasmvars->ptrb = address;
        pasmvars->ptrb0 = address;
        break;

        case 7: // ptrb[offset++]
        address = pasmvars->ptrb;
        //pasmvars->ptrb = (pasmvars->ptrb + offset) & 0x3ffff;
        pasmvars->ptrb0 = (pasmvars->ptrb + offset) & 0x3ffff;
        break;
    }

    return address;
}

void UpdatePins2(void)
{
    int32_t i;
    int32_t mask1;
    int32_t val = 0;
    int32_t mask = 0;

    for (i = 0; i < 8; i++)
    {
	if (PasmVars[i].state)
	{
	    mask1 = PasmVars[i].mem[REG_DIRA]; // dira
	    val |= mask1 & PasmVars[i].mem[REG_OUTA]; // outa
	    mask |= mask1;
	}
    }
    pin_val = (~mask) | val;
    //printf("UpdatePins2: %8.8x\n", pin_val);
}

static void SaveRegisters(PasmVarsT *pasmvars)
{
#if 0
    pasmvars->inda0 = pasmvars->inda;
    pasmvars->indb0 = pasmvars->indb;
#endif
    pasmvars->ptra0 = pasmvars->ptra;
    pasmvars->ptrb0 = pasmvars->ptrb;
}

static void UpdateRegisters(PasmVarsT *pasmvars)
{
#if 0
    pasmvars->inda = pasmvars->inda0;
    pasmvars->indb = pasmvars->indb0;
#endif
    pasmvars->ptra = pasmvars->ptra0;
    pasmvars->ptrb = pasmvars->ptrb0;
}

static int32_t FetchHubInstruction(PasmVarsT *pasmvars, int32_t prefetch)
{
    int32_t i, j, pc, lineaddr, maxnotused;

    if (prefetch)
	pc = (pasmvars->pc1 + 4) & 0xffff;
    else
	pc = pasmvars->pc & 0xffff;
    lineaddr = (pc << 2) & 0x3fff0;

    // Check if already cached
    for (i = 0; i < NUM_ICACHE; i++)
    {
        if (lineaddr == pasmvars->icachehubaddr[i]) break;
    }

    // Update usage if not prefetch
    if (!prefetch)
    {
	for (j = 0; j < NUM_ICACHE; j++)
	{
	    if (j == i)
		pasmvars->icachenotused[j] = 0;
	    else
	    {
		pasmvars->icachenotused[j]++;
		if (pasmvars->icachenotused[j] > 255)
		    pasmvars->icachenotused[j] = 255;
	    }
	}
    }

    // Return if cached
    if (i < NUM_ICACHE)
        return pasmvars->icache[i][pc&3];

    // Find the least used cache buffer
    j = 0;
    maxnotused = pasmvars->icachenotused[0];
    for (i = 1; i < NUM_ICACHE; i++)
    {
        if (pasmvars->icachenotused[i] > maxnotused)
	{
	    j = i;
	    maxnotused = pasmvars->icachenotused[i];
	}
    }

    // Fill selected cache buffer and return
#if 0
    if (!prefetch)
	printf("FetchHubInstruction: Fetch into cache %d, pc = %4.4x\n", j, pc);
    else
	printf("FetchHubInstruction: Prefetch into cache %d, pc = %4.4x\n", j, pc);
#endif
    if (!prefetch)
    {
	pasmvars->waitmode = WAIT_CACHE;
	if (nohubslots)
	    pasmvars->waitflag = 2;
	else
	    pasmvars->waitflag = ((pasmvars->cogid - loopcount) & 15) + 2;
    }
    pasmvars->icache[j][0] = LONG(lineaddr);
    pasmvars->icache[j][1] = LONG(lineaddr + 4);
    pasmvars->icache[j][2] = LONG(lineaddr + 8);
    pasmvars->icache[j][3] = LONG(lineaddr + 12);
#if 0
    pasmvars->icache[j][4] = LONG(lineaddr + 16);
    pasmvars->icache[j][5] = LONG(lineaddr + 20);
    pasmvars->icache[j][6] = LONG(lineaddr + 24);
    pasmvars->icache[j][7] = LONG(lineaddr + 28);
#endif
    pasmvars->icachehubaddr[j] = lineaddr;
    pasmvars->icachenotused[j] = 0;
    return pasmvars->icache[j][pc&3];
}

static void CheckPrefetch(PasmVarsT *pasmvars)
{
    if (!pasmvars->prefetch) return;
    if (!nohubslots && pasmvars->cogid != (loopcount & 7)) return;
    if (pasmvars->waitmode & WAIT_HUB) return;
    if (pasmvars->pc < 0x200) return;
    FetchHubInstruction(pasmvars, 1);
}

static void CheckSerialPorts(PasmVarsT *pasmvars)
{
    int addr, value1, value2;

    if (pasmvars->serina.mode)
        SerialReceive(&pasmvars->serina, pin_val);

    if (pasmvars->serinb.mode)
        SerialReceive(&pasmvars->serinb, pin_val);

    if (pasmvars->serouta.mode)
    {
        addr = ((pasmvars->serouta.pin_num >> 5) & 3) + REG_OUTA;
        value1 = pasmvars->mem[addr];
        value2 = SerialSend(&pasmvars->serouta, value1);
        if (value1 != value2)
        {
            pasmvars->mem[addr] = value2;
	    UpdatePins2();
        }
    }

    if (pasmvars->seroutb.mode)
    {
        addr = ((pasmvars->seroutb.pin_num >> 5) & 3) + REG_OUTA;
        value1 = pasmvars->mem[addr];
        value2 = SerialSend(&pasmvars->seroutb, value1);
        if (value1 != value2)
        {
            pasmvars->mem[addr] = value2;
	    UpdatePins2();
        }
    }
}

int32_t ExecutePasmInstruction2(PasmVarsT *pasmvars)
{
    int32_t cflag = pasmvars->cflag;
    int32_t zflag = pasmvars->zflag;
    int32_t instruct, pc, cond;
    int32_t opcode, value2, value1, czi;
    int32_t srcaddr, dstaddr;
    int32_t result = 0;
    int32_t i, temp, write_czr, sflag = 0, dflag = 0, psflag = 0;
    int32_t opcode_czi, rflag = 0, aflag = 0;
    int32_t returnflag = 0;
    int32_t breakflag = 0;
    int32_t hubop = 0;
    uint32_t ptr_size[12] = {0, 0, 1, 1, 2, 2, 3, 0, 1, 2, 3, 3};

    // Check serial ports
    CheckSerialPorts(pasmvars);

    // Check if multiplier working
    if (pasmvars->mulcount)
	pasmvars->mulcount--;

    // Check if we can skip further processing if not printing, not waiting
    // for a pin state and wait flag is greater than 1.
    if (pasmvars->waitflag > 1 && !pasmvars->printflag && pasmvars->waitmode != WAIT_PIN)
    {
	CheckPrefetch(pasmvars);
	pasmvars->waitflag--;
	return 0;
    }

    // Fetch a new instruction and update the pipeline
    if (!pasmvars->waitflag && pasmvars->phase == 0)
    {
        pasmvars->instruct2 = pasmvars->instruct1;
	// Update instruction pipeline and fetch new instruction
	if ((pasmvars->pc & 0xfff8) == pasmvars->dcachecogaddr)
            pasmvars->instruct1 = pasmvars->dcache[pasmvars->pc & 7];
	else if (pasmvars->pc < 0x200)
            pasmvars->instruct1 = pasmvars->mem[pasmvars->pc];
	else if (pasmvars->pc)
            pasmvars->instruct1 = FetchHubInstruction(pasmvars, 0);
	// Update PC pipeline and increment PC
        pasmvars->pc2 = pasmvars->pc1;
        pasmvars->pc1 = pasmvars->pc;
	if (pasmvars->repcnt && pasmvars->pc >= pasmvars->reptop)
	{
	    if (!pasmvars->repforever) pasmvars->repcnt--;
	    pasmvars->pc = pasmvars->repbot;
	}
	else
	    pasmvars->pc = (pasmvars->pc + 1) & 0xffff;
	// Check for REPS instruction in the second stage of the pipeline
	if ((pasmvars->instruct1 & 0xffc00000) == 0xfac00000 && !(pasmvars->pc1 & INVALIDATE_INSTR))
	{
	    pasmvars->repcnt = (pasmvars->instruct1 >> 6) & 0xffff;
	    pasmvars->repbot = pasmvars->pc;
	    pasmvars->reptop = (pasmvars->pc + (pasmvars->instruct1 & 63)) & 0xffff;
	}
	if (pasmvars->printflag) DebugPasmInstruction2(pasmvars);
        pasmvars->phase = 1;
	return 0;
    }

    // Check for cache wait
    if (pasmvars->waitflag && pasmvars->waitmode == WAIT_CACHE)
    {
	if (--pasmvars->waitflag == 0)
		pasmvars->waitmode = 0;
	else
	{
	    if (pasmvars->printflag) DebugPasmInstruction2(pasmvars);
	    return 0;
	}
    }

    // Get the instruction and pc
    instruct = pasmvars->instruct2;
    pc = pasmvars->pc2;

    // Return if instruction has been invalidated and not printing
    if (pc & INVALIDATE_INSTR)
    {
	if (!pasmvars->printflag)
	{
	    CheckPrefetch(pasmvars);
#if 0
	    if (pasmvars->waitflag)
	    {
		pasmvars->waitflag--;
		if (!pasmvars->waitflag) pasmvars->waitmode = 0;
	    }
#endif
            pasmvars->phase = 0; // Change phase to fetch mode
	    return 0;
	}
	returnflag = 1;
    }

    // Extract bit fields from the instruction
    opcode     = (instruct >> 25) & 127;
    czi        = (instruct >> 22) & 7;
    cond       = (instruct >> 18) & 15;
    dstaddr    = (instruct >> 9) & 511;
    srcaddr    = instruct & 511;
    opcode_czi = (opcode << 3) | czi;

#if 1
    // Decode the immediate flags for the source and destination fields
    if ((czi & 1) && opcode <= 0x6a)
    {
        if (opcode >= 0x61 && opcode <= 0x66)
            sflag = 3;
        else if ((opcode >= 0x4e && opcode <= 0x4f) || opcode == 0x55 || opcode == 0x60)
            sflag = 2;
        else
            sflag = 1;
    }

    dflag = ((czi & 1) && (opcode == 0x6b)) || ((czi & 2) &&
        ((opcode >= 0x59 && opcode <= 0x60) || (opcode >= 0x64 && opcode <= 0x6a)));

    // Check for extended address instructions
    if (opcode >= 0x6c)
    {
        srcaddr = instruct & 0x3ffff;
        if (opcode < 0x78)
        {
            srcaddr |= (instruct >> 4) & 0xc0000;
            sflag = (czi >> 2) + 1;
            if (opcode >= 0x78)
                dstaddr = 0x1f6 + (opcode & 3);
        }
        else
        {
            srcaddr |= (instruct >> 4) & 0x3c0000;
            sflag = 1;
        }
    }
#endif

#if 0
    // Decode the immediate flags for the source and destination fields
    sflag = (opcode <= 0x7c && (czi & 1)) || (opcode >= 0x7d && opcode <= 0x7e);

    dflag = (opcode >= 0x70 && opcode <= 0x7c && (czi & 2)) ||
            (opcode >= 0x7d && opcode <= 0x7e) ||
            (opcode == 0x7f && (instruct & 31) <= 7 && (czi & 2));

    indirect = 0;

    // Determine if ptra or ptrb are referenced
    psflag = ((opcode < 6) | (opcode_czi >= 0x340 && opcode_czi <= 0x34b)) & sflag;
    pdflag = (opcode == 0x7f && srcaddr >= 0x085 && srcaddr <= 0x087 && dflag);

    // Determine if instruction uses relative address
    rflag = (((opcode >= 0x54 && opcode <= 0x57) |
	(opcode >= 0x76 && opcode <= 0x77) |
	(opcode_czi >= 0x3d8 && opcode_czi <= 0x3e8)) & sflag) |
	(opcode_czi == 0x3ea);

    // Determine if instruction is AUGS or AUGD
    aflag = (opcode_czi >= 0x3e8 && opcode_czi <= 0x3eb);
#endif

    // Determine if ptra or ptrb are referenced
    psflag = (sflag == 3);

    // Determine if instruction uses relative address
    rflag = (sflag == 2);

    // Save ptra, ptrb, inda and indb
    SaveRegisters(pasmvars);

#if 0
    // Handle the indirect registers
    if (indirect) CheckIndRegs(pasmvars, instruct, sflag, dflag, &dstaddr, &srcaddr);
    else
#endif

    // else return if condition code is not met and
    // not using indirect registers and not augx instruction
    if (!((cond >> ((cflag << 1) | zflag)) & 1) && !aflag)
    {
	if (!pasmvars->printflag)
	{
	    CheckPrefetch(pasmvars);
#if 0
	    if (pasmvars->waitflag)
	    {
		pasmvars->waitflag--;
		if (!pasmvars->waitflag) pasmvars->waitmode = 0;
	    }
#endif
            pasmvars->phase = 0; // Change phase to fetch mode
	    return 0;
	}
	returnflag = 1;
    }

    // Set the flag that controls writing the result, zero flag and carry flag
    write_czr = czi | 1; // Assume writing result

    // Get value1 from the destination field
#if 0
    if (pdflag)
	value1 = GetPointer(pasmvars, dstaddr, 5);
    else if (dflag)
#else
    if (dflag)
#endif
    {
	if (pasmvars->augdflag)
	    value1 = (pasmvars->augdvalue << 9) | dstaddr;
	else
	    value1 = dstaddr;
    }
    else if ((dstaddr & 0x1f8) == pasmvars->dcachecogaddr)
	value1 = pasmvars->dcache[dstaddr & 7];
    else
	value1 = pasmvars->mem[dstaddr];

    // Get value2 from the source field
    //printf("psflag = %d, sflag = %d, srcaddr = %d\n", psflag, sflag, srcaddr);
    if (psflag)
    {
#if 0
	if (opcode < 6)
	    value2 = GetPointer(pasmvars, srcaddr, opcode >> 1);
	else
	    value2 = GetPointer(pasmvars, srcaddr, (opcode_czi >> 2) & 3);
#else
        value2 = GetPointer(pasmvars, srcaddr, ptr_size[(opcode_czi >> 2) - (0x61 << 1)]);
#endif
    }
    else if (sflag)
    {
	if (pasmvars->augsflag)
	    value2 = (pasmvars->augsvalue << 9) | srcaddr;
	else if (rflag)
            value2 = (srcaddr << 23) >> 23;
	else
            value2 = srcaddr;
    }
    else if (srcaddr == REG_PINA)
        value2 = pin_val;
    else if (srcaddr == REG_CNT)
        value2 = GetCnt();
    else if ((srcaddr & 0x1f8) == pasmvars->dcachecogaddr)
	value2 = pasmvars->dcache[srcaddr & 7];
    else
	value2 = pasmvars->mem[srcaddr];

    // Check sflag and dflag and reset augsflag and augdflag if needed
    if (sflag) pasmvars->augsflag = 0;
    if (dflag) pasmvars->augdflag = 0;

    // Check the wait flag if the return flag is not set
    if (returnflag)
    {
        if (pasmvars->waitflag)
	{
	    pasmvars->waitflag--;
	    if (!pasmvars->waitflag) pasmvars->waitmode = 0;
	}
    }
    else
	hubop = CheckWaitFlag2(pasmvars, instruct, value1, value2);

    if (!hubop)
	CheckPrefetch(pasmvars);

    // Print instruction if printflag set
    if (pasmvars->printflag) DebugPasmInstruction2(pasmvars);

    // Return if returnflag or waitflag is set
    if (returnflag || pasmvars->waitflag)
    {
        if (returnflag)
            pasmvars->phase = 0; // Change phase to fetch mode
         return 0;
    }

    // Check for breakpoint
    if (pc == pasmvars->breakpnt)
    {
	if (!pasmvars->printflag) pasmvars->printflag = 15;
	DebugPasmInstruction2(pasmvars);
	breakflag = 1;
    }

    // Update ptra, ptrb, inda and indb
    UpdateRegisters(pasmvars);

    // Decode the opcode and execute the instruction
    // Set the variables result, zflag and cflag.
    // Clear bits within write_czr to prevent writing variables.
    switch(opcode >> 3) // Decode the four most significant bits
    {
        case 0: // rotate and shift
	value2 &= 0x1f; // Get five LSB's
	switch (opcode)
	{
	    case 0: // ror
	    result = (((uint32_t)value1) >> value2) | (value1 << (32 - value2));
	    cflag = value1 & 1;
	    break;

	    case 1: // rol
	    result = (((uint32_t)value1) >> (32 - value2)) | (value1 << value2);
	    cflag = (value1 >> 31) & 1;
	    break;

	    case 2: // shr
	    result = (((uint32_t)value1) >> value2);
	    cflag = value1 & 1;
	    break;

	    case 3: // shl
	    result = (value1 << value2);
	    cflag = (value1 >> 31) & 1;
	    break;

	    case 4: // rcr
	    if (value2)
	    {
	        result = (cflag << 31) | (((uint32_t)value1) >> 1);
		result >>= (value2 - 1);
	    }
	    else
	        result = value1;
	    cflag = value1 & 1;
	    break;

	    case 5: // rcl
	    result = cflag ? (1 << value2) - 1 : 0;
	    result |= (value1 << value2);
	    cflag = (value1 >> 31) & 1;
	    break;

	    case 6: // sar
	    result = value1 >> value2;
	    cflag = value1 & 1;
	    break;

            case 7: // sal
	    NotImplemented(instruct);
            break;
	}
	zflag = (result == 0);
	break;

        case 1: // addxx, subxx
	switch (opcode & 7)
	{
	    case 0: // add
	    result = value1 + value2;
	    cflag = (((value1 & value2) | ((value1 | value2) & (~result))) >> 31) & 1;
	    break;

	    case 1: // addx
	    result = value1 + value2 + cflag;
	    cflag = (((value1 & value2) | ((value1 | value2) & (~result))) >> 31) & 1;
	    zflag = (result == 0) & zflag;
	    break;

	    case 2: // adds
	    result = value1 + value2;
	    cflag = (((~(value1 ^ value2)) & (value1 ^ result)) >> 31) & 1;
	    zflag = (result == 0);
	    break;

	    case 3: // addsx
	    result = value1 + value2 + cflag;
	    cflag = (((~(value1 ^ value2)) & (value1 ^ result)) >> 31) & 1;
	    zflag = (result == 0) & zflag;
	    break;

	    case 4: // sub
	    result = value1 - value2;
            //printf("result = %d, value1 = %d, value2 = %d\n", result, value1, value2);
	    cflag = ((uint32_t)value1) < ((uint32_t)value2);
	    break;

	    case 5: // subx
	    result = value1 - value2 - cflag;
	    if (value2 != 0xffffffff || !cflag)
	        cflag = ((uint32_t)value1) < ((uint32_t)(value2 + cflag));
	    zflag = (result == 0) & zflag;
	    break;

	    case 6: // subs
	    result = value1 - value2;
	    zflag = (result == 0);
	    cflag = (((value1 ^ value2) & (value1 ^ result)) >> 31) & 1;
	    break;

	    case 7: // subsx
	    result = value1 - value2 - cflag;
	    cflag = (((value1 ^ value2) & (value1 ^ result)) >> 31) & 1;
	    zflag = (result == 0) & zflag;
	    break;
	}
	zflag = (result == 0);
	break;

        case 2: // cmpxx, subr, cmpsub
	switch (opcode & 7)
	{
	    case 0: // cmp
            //printf("\ncmp $%x $%x\n", value1, value2);
	    result = value1 - value2;
	    zflag = (result == 0);
	    cflag = ((uint32_t)value1) < ((uint32_t)value2);
            write_czr &= 6;
	    break;

	    case 1: // cmpx
	    result = value1 - value2 - cflag;
	    if (value2 != 0xffffffff || !cflag)
	        cflag = ((uint32_t)value1) < ((uint32_t)(value2 + cflag));
	    zflag = (result == 0) & zflag;
            write_czr &= 6;
	    break;

	    case 2: // cmps
            //printf("\ncmps $%x $%x\n", value1, value2);
	    result = value1 - value2;
	    zflag = (result == 0);
	    cflag =  value1 < value2;
            write_czr &= 6;
	    break;

	    case 3: // cmpsx
	    result = value1 - value2 - cflag;
	    cflag = (((value1 ^ value2) & (value1 ^ result)) >> 31) & 1;
	    zflag = (result == 0) & zflag;
            write_czr &= 6;
	    break;

	    case 4: // cmpr
	    result = value2 - value1;
	    zflag = (result == 0);
	    cflag = ((uint32_t)value2) < ((uint32_t)value1);
            write_czr &= 6;
	    break;

	    case 5: // cmpm
            NotImplemented(instruct);
            write_czr &= 6;
            break;

	    case 6: // subr
	    result = value2 - value1;
	    cflag = ((uint32_t)value2) < ((uint32_t)value1);
	    zflag = (result == 0);
            break;

	    case 7: // cmpsub
	    cflag = (((uint32_t)value1) >= ((uint32_t)value2));
	    result = cflag ? value1 - value2 : value1;
	    zflag = (result == 0);
            break;
	}
	break;

        case 3: // min, max, mins, maxs, sumxx
        switch (opcode & 7)
        {
	    case 0: // min
	    cflag = (((uint32_t)value1) < ((uint32_t)value2));
	    zflag = (value2 == 0);
	    result = cflag ? value2 : value1;
	    break;

	    case 1: // max
	    cflag = (((uint32_t)value1) < ((uint32_t)value2));
	    zflag = (value2 == 0);
	    result = cflag ? value1 : value2;
	    break;

	    case 2: // mins
	    cflag = (value1 < value2);
	    zflag = (value2 == 0);
	    result = cflag ? value2 : value1;
	    break;

	    case 3: // maxs
	    cflag = (value1 < value2);
	    zflag = (value2 == 0);
	    result = cflag ? value1 : value2;
	    break;

	    case 4: // sumc
	    result = cflag ? value1 - value2 : value1 + value2;
	    cflag = (~cflag) << 31;
	    cflag = (((cflag ^ value1 ^ value2) & (value1 ^ result)) >> 31) & 1;
	    break;

	    case 5: // sumnc
	    result = cflag ? value1 + value2 : value1 - value2;
	    cflag = cflag << 31;
	    cflag = (((cflag ^ value1 ^ value2) & (value1 ^ result)) >> 31) & 1;
	    break;

	    case 6: // sumz
	    result = zflag ? value1 - value2 : value1 + value2;
	    cflag = (~zflag) << 31;
	    cflag = (((cflag ^ value1 ^ value2) & (value1 ^ result)) >> 31) & 1;
	    break;

	    case 7: // sumnz
	    result = zflag ? value1 + value2 : value1 - value2;
	    cflag = zflag << 31;
	    cflag = (((cflag ^ value1 ^ value2) & (value1 ^ result)) >> 31) & 1;
	    break;

        }
	break;

        case 4: // isob, notb, clrb, setbxx
	value2 = (value2 & 0x1f);
	cflag = (value1 >> value2) & 1;
        switch ((value2 >> 5) & 7)
        {
	    case 0: // isob
	    result = (value1 >> value2) & 1;
	    break;

	    case 1: // notb
	    result = value1 ^ (1 << value2);
	    break;

	    case 2: // clrb
	    result = value1 & (~(1 << value2));
	    break;

	    case 3: // setb
	    result = value1 | (1 << value2);
	    break;

	    case 4: // setbc
	    result = value1 & (~(1 << value2));
	    result |= cflag << value2;
	    break;

	    case 5: // setbnc
	    result = value1 & (~(1 << value2));
	    result |= (cflag ^ 1) << value2;
	    break;

	    case 6: // setbz
	    result = value1 & (~(1 << value2));
	    result |= zflag << value2;
	    break;

	    case 7: // setbnz
	    result = value1 & (~(1 << value2));
	    result |= (zflag ^ 1) << value2;
	    break;
	}
	zflag = (result == 0);
        break;

        case 5: // andn, and, or, xor, muxxx
	switch (opcode & 7)
	{
	    case 0: // andn
	    result = value1 & (~value2);
	    break;

	    case 1: // and
	    result = value1 & value2;
	    break;

	    case 2: // or
	    result = value1 | value2;
	    break;

	    case 3: // xor
	    result = value1 ^ value2;
	    break;

	    case 4: // muxc
	    result = (value1 & (~value2)) | (value2 & (-cflag));
	    break;

	    case 5: // muxnc
	    result = (value1 & (~value2)) | (value2 & (~(-cflag)));
	    break;

	    case 6: // muxz
	    result = (value1 & (~value2)) | (value2 & (-zflag));
	    break;

	    case 7: // muxnz
	    result = (value1 & (~value2)) | (value2 & (~(-zflag)));
	    break;
	}
	zflag = (result == 0);
	cflag = parity(result);
	break;

        case 6: // mov, not abs, negxx
	switch (opcode & 7)
	{
	    case 0: // mov
	    result = value2;
	    cflag = (value2 >> 31) & 1;
	    break;

	    case 1: // not
	    cflag = value2 < 0;
	    result = ~value2;
	    break;

	    case 2: // abs
	    cflag = (value2 >> 31) & 1;
	    result = _abs(value2);
	    break;

	    case 3: // neg
	    cflag = (value2 >> 31) & 1;
	    result = -value2;
	    break;

	    case 4: // negc
	    result = cflag ? -value2 : value2;
	    cflag = (value2 >> 31) & 1;
	    break;

	    case 5: // negnc
	    result = cflag ? value2 : -value2;
	    cflag = (value2 >> 31) & 1;
	    break;

	    case 6: // negz
	    result = zflag ? -value2 : value2;
	    cflag = (value2 >> 31) & 1;
	    break;

	    case 7: // negnz
	    result = zflag ? value2 : -value2;
	    cflag = (value2 >> 31) & 1;
	    break;
	}
	zflag = (result == 0);
	break;

        case 7: // altds, decod, topone, botone, incmod, decmod, mul, muls
	switch (opcode & 7)
	{
	    case 0: // altds
	    NotImplemented(instruct);
	    break;

	    case 1: // decod
	    result = 1 << (value1 & 0x1f);
	    break;

	    case 2: // topone
            for (result = 31; result >= 0; result--)
            {
                if (value2 & 0x80000000) break;
                value2 <<= 1;
            }
	    break;

	    case 3: // botone
            for (result = 0; result <= 31; result++)
            {
                if (value2 & 1) break;
                value2 >>= 1;
            }
	    break;

            case 4: // incmod
	    cflag = (value1 == value2);
	    if (cflag)
	        result = 0;
	    else
	        result = value1 + 1;
	    break;

            case 5: // decmod
	    cflag = (value1 == 0);
	    if (cflag)
	        result = value2;
	    else
	        result = value1 - 1;
	    break;

	    case 6: // mul
	    value1 &= 0xffff;
	    value2 &= 0xffff;
	    result = value1 * value2;
            break;

	    case 7: // muls
	    value1 = (value1 << 16) >> 16;
	    value2 = (value2 << 16) >> 16;
	    result = value1 * value2;
            break;
	}
	zflag = (result == 0);
	break;

        case 8: // setnibn, getnibn, rolnibn, setbytn, getbytn
	write_czr = 1;
	temp = (instruct >> 21) & 0x1c;
        switch(opcode & 7)
        {
            case 0: // setnibn
            case 1: // setnibn
	    result = (value2 & 15) << temp;
	    temp = ~(15 << temp);
	    result |= (value1 & temp);
            break;

            case 2: // getnibn
            case 3: // getnibn
	    result = (value2 >> temp) & 15;
            break;

            case 4: // rolnibn
            case 5: // rolnibn
	    result = (value2 << temp) | ((uint32_t)value2 >> (32 - temp));
            break;

	    case 6: // setbytn
	    write_czr = 1;
	    result = (value2 & 0xff) << temp;
	    temp = ~(0xff << temp);
	    result |= (value1 & temp);
	    break;

	    case 7: // getbytn
	    write_czr = 1;
	    result = (value2 >> temp) & 0xff;
	    break;
	}
	break;

        case 9: // rolxxx, setxxx, getxxx, djxx, jxx, movbyts, splitx, mergeb, rev
        switch(opcode&7)
        {
            case 0: // rolbytn
            NotImplemented(instruct);
            break;

            case 1: // setwrdn, getwrdn
            NotImplemented(instruct);
            break;

            case 2: // rolwrdn, setbyts, movbyts
            NotImplemented(instruct);
            break;

            case 3: // splitb, mergeb, splitw, mregew
            switch(czi >> 1)
            {
                case 0: // splitb
                NotImplemented(instruct);
                break;

                case 1: // mergeb
                NotImplemented(instruct);
                break;

	        case 2: // splitw
	        for (i = 0; i < 16; i++)
	        {
		    result = (result << 1) |
		        ((value1 >> 15)&0x10000) | ((value1 >> 30)&1);
		    value1 <<= 1;
	        }
	        break;

	        case 3: // mergew
	        for (i = 0; i < 16; i++)
	        {
		    result = (result << 2) |
		        ((value1 >> 30) & 2) | ((value1 >> 15) & 1);
		    value1 <<= 1;
	        }
	        break;
            }

            case 4: // sets, gets, setd, getd
            switch (czi >> 1)
            {
	        case 0: // sets
	        result = (value1 & 0xfffffe00) | (value2 & 0x1ff);
	        break;

	        case 1: // gets
	        result = (value2 &0x1ff);
	        break;

	        case 2: // setd
	        result = (value1 & 0xfffc01ff) | ((value2 & 0x1ff) << 9);
	        break;

	        case 3: // getd
	        result = ((value2 >> 9) & 0x1ff);
	        break;
            }
            break;

            case 5: // setds, setcond, seti, rev
            switch (czi >> 1)
            {
	        case 0: // setds
	        result = (value1 & 0xfffc0000) | (value2 & 0x3ffff);
	        break;

	        case 1: // setcond
	        result = (value1 & 0xff83ffff) | ((value2 & 0x1f) << 18);
	        break;

	        case 2: // seti
	        result = (value1 & 0x007fffff) | ((value2 & 0x1ff) << 23);
	        break;

	        case 3: // rev
	        cflag = value1 & 1;
	        value2 = 32 - value2;
	        result = 0;
	        while (value2-- > 0)
	        {
		    result = (result << 1) | (value1 & 1);
		    value1 >>= 1;
	        }
	        break;
            }
            break;

	    case 6: // djz, djnz, djs, djns
            value1--;
	    cflag = (value1 == -1);
	    case 7: // jz, jnz, js, jns
            result = value1;
	    zflag = (result == 0);
	    // Determine if we should jump
            if (czi & 4) // djs, djns, js, jns
            {
                NotImplemented(instruct);
            }
            else // djz, djnz, jz, jnz
            {
	        if (zflag != ((czi >> 1) & 1))
	        {
                    value2 = (value2 << 23) >> 23;
	            pasmvars->pc = (pc + value2) & 0x1ffff;
                    pasmvars->pc1 |= INVALIDATE_INSTR;
	        }
            }
	    break;
	}
	zflag = (result == 0);
        break;

        case 10: // testx, anyb, waitcnt
        write_czr &= 6;
        switch(opcode & 7)
        {
	    case 0: // testn
	    result = value1 & (~value2);
	    zflag = (result == 0);
	    cflag = parity(result);
	    break;

	    case 1: // test
	    result = value1 & value2;
	    zflag = (result == 0);
	    cflag = parity(result);
	    break;

            case 2: // anyb
	    result = value1 | value2;
	    zflag = (result == 0);
	    cflag = parity(result);
            break;

	    case 3: // testb
	    value2 = (value2 & 0x1f);
	    cflag = (value1 >> value2) & 1;
	    zflag = cflag;
	    break;

            case 4: // waitcnt
	    if (value1 != GetCnt())
	    {
		printf("ERROR: waitcnt value1 = %8.8x, cnt = %8.8x\n", value1, GetCnt());
		spinsim_exit(1);
	    }
	    result = value1 + value2;
	    zflag = (result == 0);
	    cflag = (((value1 & value2) | ((value1 | value2) & (~result))) >> 31) & 1;
            write_czr |= 1;
            break;

            case 5: // calld
            if (rflag)
                 pasmvars->pc = (pc + value2) & 0xffff;
            else
                 pasmvars->pc = value2 & 0xffff;
            pasmvars->pc1 |= INVALIDATE_INSTR;
            result = (pc + 1) | (pasmvars->zflag << 17) | (pasmvars->cflag << 16);
            write_czr |= 1;
            break;

            case 6: // rdlut
            NotImplemented(instruct);
            write_czr |= 1;
            break;

            case 7: // msgin
            NotImplemented(instruct);
            write_czr |= 1;
            break;
	}
	break;
            
        case 11: // wrlut, msgout, rep, xinit, xzero, xcont, waitxxx
        switch(opcode&7)
        {
	    case 1: // wrlut, msgout
	    NotImplemented(instruct);
	    break;

            case 2: // rep, xinit
            if (czi & 4) // xinit
		NotImplemented(instruct);
            else // rep
            {
                pasmvars->repcnt = value2;
                pasmvars->repbot = pasmvars->pc;
                pasmvars->reptop = (pasmvars->pc + value1) & 0xffff;
            }
	    break;

            case 3: // xzero, xcont
            NotImplemented(instruct);
            break;

            case 4: // waitpae
            NotImplemented(instruct);
            break;

            case 5: // waitpan
            NotImplemented(instruct);
            break;

            case 6: // waitpbe
            NotImplemented(instruct);
            break;

            case 7: // waitpbn
            NotImplemented(instruct);
            break;
	}
	break;

	case 12: // jp, jnp, rdxxxx, wrxxxx, fblock, coginit
	switch(opcode & 7)
	{
            case 0: // jp, jnp
            if (czi & 4) // jn
            {
	        if (value1 < 32 && !((pin_val >> value1) & 1))
	        {
		    if (instruct & 0x10000)
		        pasmvars->pc = (pc + value2) & 0xffff;
		    else
	                pasmvars->pc = value2 & 0xffff;
	        }
            }
            else // jp
            {
	        if (value1 < 32 && ((pin_val >> value1) & 1))
	        {
		    if (instruct & 0x10000)
		        pasmvars->pc = (pc + value2) & 0xffff;
		    else
	                pasmvars->pc = value2 & 0xffff;
	        }
            }
            break;
 
	    case 1: // rdbyte
	    result = BYTE(value2);
            if (pasmvars->printflag > 1)
	        fprintf(tracefile, ", rdb[%x]", value2);
            break;

	    case 2: // rdword
	    result = WORD(value2);
            if (pasmvars->printflag > 1)
	        fprintf(tracefile, ", rdw[%x]", value2);
            break;

	    case 3: // rdlong
	    result = LONG(value2);
            if (pasmvars->printflag > 1)
	        fprintf(tracefile, ", rdl[%x]", value2);
            break;

            case 4: // rdfast, wrbyte
            if (czi & 4) // wrbyte
            {
                write_czr &= 6;
	        BYTE(value2) = value1;
                if (pasmvars->printflag > 1)
	            fprintf(tracefile, ", wrb[%x] = %x", value2, value1);
            }
            else // rdfast
            {
                NotImplemented(instruct);
            }
            break;

            case 5: // wrword, wrlong
            write_czr &= 6;
            if (czi & 4) // wrlong
            {
	        LONG(value2) = value1;
                if (pasmvars->printflag > 1)
	            fprintf(tracefile, ", wrl[%x] = %x", value2, value1);
            }
            else // wrword
            {
	        WORD(value2) = value1;
                if (pasmvars->printflag > 1)
	            fprintf(tracefile, ", wrw[%x] = %x", value2, value1);
            }
            break;

            case 6: // wrfast, fblock
            if (czi & 4) // fblock
            {
                NotImplemented(instruct);
            }
            else // wrfast
            {
                NotImplemented(instruct);
            }
            break;

	    case 7: // coginit
            NotImplemented(instruct);
            break;
	}
	zflag = (result == 0);
	break;

        case 13: // qxxx, jmp, call, calla, callb, misc
        switch(opcode & 7)
        {
	    case 0: // qmul, qdiv
	    NotImplemented(instruct);
            break;

	    case 1: // qsqr, qsin
	    NotImplemented(instruct);
            break;

	    case 2: // qrot, qatn
	    NotImplemented(instruct);
            break;

	    case 3: // misc
            switch (srcaddr)
            {
                case 0: // clkset
                NotImplemented(instruct);
                break;

                case 1: // cogid
                result = pasmvars->cogid;
                break;

                case 3: // cogstop
		for (result = 0; result < 8; result++)
		{
		    if (!PasmVars[result].state) break;
		}
		cflag = (result == 8);
		result = value1 & 7;
		zflag = (result == 0);
                PasmVars[result].state = 0;
		UpdatePins2();
		// Return without saving if we stopped this cog
		if (result == pasmvars->cogid) return breakflag;
		break;

                case 4: // locknew
                NotImplemented(instruct);
                break;

                case 5: // lockret
                NotImplemented(instruct);
                break;

                case 6: // lockclr
                NotImplemented(instruct);
                break;

                case 7: // lockset
                NotImplemented(instruct);
                break;

                case 14: // qlog
                NotImplemented(instruct);
                break;

                case 15: // qexp
                NotImplemented(instruct);
                break;

                case 17: // rfbyte
                NotImplemented(instruct);
                break;

                case 18: // rfword
                NotImplemented(instruct);
                break;

                case 19: // rflong
                NotImplemented(instruct);
                break;

                case 20: // wfmask
                NotImplemented(instruct);
                break;

                case 21: // wfbyte
                NotImplemented(instruct);
                break;

                case 22: // wfword
                NotImplemented(instruct);
                break;

                case 23: // wflong
                NotImplemented(instruct);
                break;

                case 24: // setdac
                NotImplemented(instruct);
                break;

                case 25: // setfrq
                NotImplemented(instruct);
                break;

                case 26: // getro
                NotImplemented(instruct);
                break;

                case 27: // getbw
                NotImplemented(instruct);
                break;

                case 28: // getgx
                NotImplemented(instruct);
                break;

                case 29: // getgy
                NotImplemented(instruct);
                break;

                case 30: // getcnt
                result = GetCnt();
                break;

                case 31: // getrnd
                NotImplemented(instruct);
                break;

                case 32: // setq
                NotImplemented(instruct);
                break;

                case 34: // getqx
                NotImplemented(instruct);
                break;

                case 35: // getqy
                NotImplemented(instruct);
                break;

                case 36: // waitx
                NotImplemented(instruct);
                break;

                case 37: // waitpx
                NotImplemented(instruct);
                break;

                case 38: // waitpr
                NotImplemented(instruct);
                break;

                case 39: // waitpf
                NotImplemented(instruct);
                break;

                case 40: // push
		pasmvars->retstack[pasmvars->retptr] = value1;
		pasmvars->retptr = (pasmvars->retptr + 1) & 3;
		if (pasmvars->retptr == 0)
		    printf("return stack overflow\n");
                break;

                case 41: // call
                if (instruct & 0x20000)
                     pasmvars->pc = (pc + instruct) & 0xffff;
                else
                    pasmvars->pc = instruct & 0xffff;
                pasmvars->pc1 |= INVALIDATE_INSTR;
                pc++;
                pc |= (pasmvars->zflag << 17) | (pasmvars->cflag << 16);
                pasmvars->retstack[pasmvars->retptr] = pc;
                pasmvars->retptr = (pasmvars->retptr + 1) & 3;
                if (pasmvars->retptr == 0)
                    printf("return stack overflow\n");
                break;

                case 42: // calla
                if (instruct & 0x20000)
                     pasmvars->pc = (pc + instruct) & 0xffff;
                else
                     pasmvars->pc = instruct & 0xffff;
                     pasmvars->pc1 |= INVALIDATE_INSTR;
                pc++;
                pc |= (pasmvars->zflag << 17) | (pasmvars->cflag << 16);
                LONG(pasmvars->ptra) = pc;
                pasmvars->ptra = (pasmvars->ptra + 4) & 0x3ffff;
                break;

                case 43: // callb
                if (instruct & 0x20000)
                     pasmvars->pc = (pc + instruct) & 0xffff;
                else
                     pasmvars->pc = instruct & 0xffff;
                     pasmvars->pc1 |= INVALIDATE_INSTR;
                pc++;
                pc |= (pasmvars->zflag << 17) | (pasmvars->cflag << 16);
                LONG(pasmvars->ptrb) = pc;
                pasmvars->ptrb = (pasmvars->ptrb + 4) & 0x3ffff;
                break;

                case 44: // pop
		pasmvars->retptr = (pasmvars->retptr - 1) & 3;
		pasmvars->retstack[pasmvars->retptr] = value1;
		if (pasmvars->retptr == 3)
		    printf("return stack underflow\n");
                break;

                case 45: // ret
                write_czr &= 6;
		pasmvars->retptr = (pasmvars->retptr - 1) & 3;
		result = pasmvars->retstack[pasmvars->retptr];
		if (pasmvars->retptr == 3)
		    printf("return stack underflow\n");
		cflag = (result >> 16) & 1;
		zflag = (result >> 17) & 1;
		pasmvars->pc = result & 0xffff;
                pasmvars->pc1 |= INVALIDATE_INSTR;
                break;

                case 46: // reta
                write_czr &= 6;
		pasmvars->ptra = (pasmvars->ptra - 4) & 0x3ffff;
		result = LONG(pasmvars->ptra);
		cflag = (result >> 16) & 1;
		zflag = (result >> 17) & 1;
		pasmvars->pc = result & 0xffff;
                pasmvars->pc1 |= INVALIDATE_INSTR;
                break;

                case 47: // retb
                write_czr &= 6;
		pasmvars->ptrb = (pasmvars->ptrb - 4) & 0x3ffff;
		result = LONG(pasmvars->ptrb);
		cflag = (result >> 16) & 1;
		zflag = (result >> 17) & 1;
		pasmvars->pc = result & 0xffff;
                pasmvars->pc1 |= INVALIDATE_INSTR;
                break;

                default:
                NotImplemented(instruct);
                break;
            }
            break;

            case 4: // jmp
            if (rflag)
                pasmvars->pc = (pc + instruct) & 0xffff;
            else
                pasmvars->pc = instruct & 0xffff;
            pasmvars->pc1 |= INVALIDATE_INSTR;
            write_czr &= 6;
            break;

            case 5: // call
            if (rflag)
                pasmvars->pc = (pc + instruct) & 0xffff;
            else
                pasmvars->pc = instruct & 0xffff;
            pasmvars->pc1 |= INVALIDATE_INSTR;
            pc++;
            pc |= (pasmvars->zflag << 17) | (pasmvars->cflag << 16);
            pasmvars->retstack[pasmvars->retptr] = pc;
            pasmvars->retptr = (pasmvars->retptr + 1) & 3;
            if (pasmvars->retptr == 0)
                printf("return stack overflow\n");
            write_czr &= 6;
            break;

            case 6: // calla
            if (rflag)
                 pasmvars->pc = (pc + instruct) & 0xffff;
            else
                 pasmvars->pc = instruct & 0xffff;
                 pasmvars->pc1 |= INVALIDATE_INSTR;
            pc++;
            pc |= (pasmvars->zflag << 17) | (pasmvars->cflag << 16);
            LONG(pasmvars->ptra) = pc;
            pasmvars->ptra = (pasmvars->ptra + 4) & 0x3ffff;
            write_czr &= 6;
            break;

            case 7: // callb
            if (rflag)
                 pasmvars->pc = (pc + instruct) & 0xffff;
            else
                 pasmvars->pc = instruct & 0xffff;
                 pasmvars->pc1 |= INVALIDATE_INSTR;
            pc++;
            pc |= (pasmvars->zflag << 17) | (pasmvars->cflag << 16);
            LONG(pasmvars->ptrb) = pc;
            pasmvars->ptrb = (pasmvars->ptrb + 4) & 0x3ffff;
            write_czr &= 6;
            break;
        }
        break;

        case 14: // calld, loc
        if (opcode & 4) // loc
        {
            if (instruct & 0x20000)
                result = (pc + value2) << 2;
            else
                result = value2 << 2;
            result &= 0x3ffff;
        }
        else // calld
        {
	    NotImplemented(instruct);
        }
	break;

        case 15: // augs, augd
        if (opcode & 4) // augd
        {
            pasmvars->augdflag = 1;
            pasmvars->augdvalue = instruct & 0x007fffff;
        }
        else // augs
        {
            pasmvars->augsflag = 1;
            pasmvars->augsvalue = instruct & 0x007fffff;
        }
        break;
    }

    // Conditionally update flags and write result
    if (write_czr & 2)
    {
	pasmvars->zflag = zflag;
        if (pasmvars->printflag > 1) fprintf(tracefile, ", z = %d", zflag);
    }
    if (write_czr & 4)
    {
	pasmvars->cflag = cflag;
        if (pasmvars->printflag > 1) fprintf(tracefile, ", c = %d", cflag);
    }
    if (write_czr & 1)
    {
        pasmvars->lastd = result;
	if ((dstaddr & 0x1f8) == pasmvars->dcachecogaddr)
	{
	    pasmvars->dcache[dstaddr & 7] = result;
            if (pasmvars->printflag > 1)
	      fprintf(tracefile, ", dcache[%d] = %x", dstaddr & 7, result);
	}
	else
	{
	    pasmvars->mem[dstaddr] = result;
            if (pasmvars->printflag > 1)
	        fprintf(tracefile, ", cram[%x] = %x", dstaddr, result);
	}
	// Check if we need to update the pins
	if (dstaddr >= REG_OUTA && dstaddr <= REG_DIRB) UpdatePins2();
    }
    if (pasmvars->waitflag)
    {
        fprintf(tracefile, "XXXXXXXXXX BAD XXXXXXXXXXXXXXX\n");
	pasmvars->waitflag--;
	if (!pasmvars->waitflag) pasmvars->waitmode = 0;
    }

    pasmvars->phase = 0; // Change phase to fetch mode

    return breakflag;
}
/*
+------------------------------------------------------------------------------------------------------------------------------+
|                                                   TERMS OF USE: MIT License                                                  |
+------------------------------------------------------------------------------------------------------------------------------+
|Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation    |
|files (the "Software"), to deal in the Software without restriction, including without limitation the rights to use, copy,    |
|modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software|
|is furnished to do so, subject to the following conditions:                                                                   |
|                                                                                                                              |
|The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.|
|                                                                                                                              |
|THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE          |
|WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR         |
|COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,   |
|ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.                         |
+------------------------------------------------------------------------------------------------------------------------------+
*/
